__all__ = ["to_dict", "AppDbDocument", "AppDbCollection", "AppDbCollections"]

import asyncio
import datetime as dt
import numbers
from dataclasses import dataclass, field
from datetime import datetime
from typing import Callable, list

import httpx

from ..auth import DomoAuth
from ..base.entities import DomoEntity
from ..routes import appdb as appdb_routes
from ..utils import (
    chunk_execution as dmce,
    convert as dlcv,
)


def to_dict(value):
    if hasattr(value, "to_dict"):
        return value.to_dict()

    if isinstance(value, dict):
        return {key: to_dict(v) for key, v in value.items()}

    if isinstance(value, list):
        return [to_dict(v) for v in value]

    if isinstance(value, numbers.Number):
        return value

    return str(value)


@dataclass
class AppDbDocument(DomoEntity):
    auth: DomoAuth | None = field(repr=False)
    _id: str | None = None
    _created_on_dt: datetime | None = None
    _updated_on_dt: datetime | None = None
    content: dict | None = None
    _collection_id: str | None = None
    _identity_columns: list[str] | None = None

    def to_dict(self, custom_content_to_dict_fn: Callable | None = None):
        self.update_config()

        s = {"id": self._id, "collectionId": self._collection_id}

        if custom_content_to_dict_fn:
            s.update({"content": custom_content_to_dict_fn(self.content)})

        else:
            for key, value in self.__dict__.items():
                if key.startswith("_") or key in ["auth"]:
                    continue

                s.update({key: to_dict(value)})

        return s

    def __eq__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False

        if self._identity_columns:
            return all(
                getattr(self, col) == getattr(other, col)
                for col in self._identity_columns
            )

        return self._id == other._id

    @classmethod
    def from_dict(
        cls,
        auth: DomoAuth,
        content,
        new_cls,
        identity_columns,
        collection_id=None,
        document_id=None,
        metadata=None,
        created_on_dt=None,
        updated_on_dt=None,
    ):
        if metadata:
            collection_id = metadata.pop("collectionId")

            created_on_dt = dlcv.convert_string_to_datetime(metadata.pop("createdOn"))

            updated_on_dt = dlcv.convert_string_to_datetime(metadata.pop("updatedOn"))
            document_id = metadata["id"]

        return cls(
            auth=auth,
            _id=document_id,
            _created_on_dt=created_on_dt,
            _updated_on_dt=updated_on_dt,
            content=content,
            _collection_id=collection_id,
            _identity_columns=identity_columns or [],
        )

    @classmethod
    async def create_document(
        cls,
        content: dict,
        collection_id: str,
        auth: DomoAuth,
        session: httpx.AsyncClient,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
    ):
        res = await appdb_routes.create_document(
            auth=auth,
            collection_id=collection_id,
            content=content,
            session=session,
            debug_api=debug_api,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return await cls.get_by_id(
            collection_id=collection_id, document_id=res.response["id"], auth=auth
        )

    async def update_document(
        self,
        content: dict | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        debug_num_stacks_to_drop=1,
        return_raw: bool = False,
    ):
        res = await appdb_routes.update_document(
            auth=self.auth,
            collection_id=self._collection_id,
            document_id=self._id,
            content=content or self.to_dict()["content"],
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
            parent_class=self.__class__.__name__,
        )

        if return_raw:
            return res

        return await AppDbDocument.get_by_id(
            collection_id=self._collection_id, document_id=self._id, auth=self.auth
        )

    @classmethod
    async def upsert(
        cls,
        auth: DomoAuth,
        collection_id,
        content: dict,
        identity_columns: list[str],
        session: httpx.AsyncClient | None = None,
        debug_api=False,
        debug_num_stacks_to_drop=3,
        return_raw: bool = False,
    ):
        domo_doc = None

        {f"content.{col}": content[col] for col in identity_columns}

        domo_collection = await AppDbCollection.get_by_id(
            auth=auth, collection_id=collection_id, return_raw=False
        )

        if domo_collection.domo_documents:
            domo_doc = next(
                (
                    doc
                    for doc in domo_collection.domo_documents
                    if all(
                        [
                            doc.content.get(col) == content.get(col)
                            for col in identity_columns
                        ]
                    )
                ),
                None,
            )

        if domo_doc:
            return await domo_doc.update_document(
                content=content,
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                session=session,
                return_raw=return_raw,
            )

        return await cls.create_document(
            content=content,
            collection_id=collection_id,
            auth=auth,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
            return_raw=return_raw,
        )

    @classmethod
    def _from_api(
        cls,
        auth: DomoAuth,
        obj,
        identity_columns: list[str] = None,
    ):
        content = obj.pop("content")

        return cls.from_dict(
            auth=auth,
            content=content,
            new_cls=cls,
            identity_columns=identity_columns,
            metadata=obj,
        )

    @classmethod
    def from_json(
        cls,
        auth: DomoAuth,
        collection_id: str,
        content: dict,
        identity_columns: list[str] = None,
    ):
        return cls.from_dict(
            auth=auth,
            content=content,
            new_cls=cls,
            identity_columns=identity_columns,
            collection_id=collection_id,
        )

    def update_config(self):
        self.content = {
            key: value
            for key, value in self.__dict__.items()
            if key not in ["auth", "content"] and not key.startswith("_")
        }
        return self.content

    @classmethod
    async def get_by_id(
        cls,
        collection_id: str,
        document_id: str,
        auth: DomoAuth,
        identity_columns=None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=1,
        return_raw: bool = False,
    ):
        res = await appdb_routes.get_collection_document_by_id(
            auth=auth,
            collection_id=collection_id,
            document_id=document_id,
            parent_class=cls.__name__,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return cls._from_api(
            auth=auth,
            obj=res.response,
            identity_columns=identity_columns or [],
        )


@dataclass
async def create_document(
    cls,
    content: dict,
    collection_id: str,
    auth: DomoAuth,
    session: httpx.AsyncClient,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    return_raw: bool = False,
):
    res = await appdb_routes.create_document(
        auth=auth,
        collection_id=collection_id,
        content=content,
        session=session,
        debug_api=debug_api,
        parent_class=cls.__name__,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    return await cls.get_by_id(
        collection_id=collection_id, document_id=res.response["id"], auth=auth
    )


async def update_document(
    self,
    content: dict = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
    return_raw: bool = False,
):
    res = await appdb_routes.update_document(
        auth=self.auth,
        collection_id=self._collection_id,
        document_id=self._id,
        content=content or self.to_dict()["content"],
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res

    return res


@dataclass
class AppDbCollection(DomoEntity):
    auth: DomoAuth = field(repr=False)
    id: str
    name: str

    created_on_dt: dt.datetime
    updated_on_dt: dt.datetime

    schema: dict

    domo_documents: list[AppDbDocument] = None

    @classmethod
    def from_dict(cls, auth, obj):
        return cls(
            auth=auth,
            id=obj["id"],
            name=obj["name"],
            created_on_dt=dlcv.convert_string_to_datetime(obj["createdOn"]),
            updated_on_dt=dlcv.convert_string_to_datetime(obj["updatedOn"]),
            schema=obj["schema"],
        )

    @classmethod
    async def get_by_id(
        cls,
        auth: DomoAuth,
        collection_id,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
    ):
        res = await appdb_routes.get_collection_by_id(
            auth=auth,
            collection_id=collection_id,
            parent_class=cls.__name__,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return cls.from_dict(auth=auth, obj=res.response)

    async def share_collection(
        self,
        domo_user=None,
        domo_group=None,
        permission: appdb_routes.Collection_Permission_Enum = appdb_routes.Collection_Permission_Enum.READ_CONTENT,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient = None,
    ):
        return await appdb_routes.modify_collection_permissions(
            collection_id=self.id,
            user_id=(domo_user and domo_user.id)
            or (await self.auth.who_am_i()).response["id"],
            group_id=domo_group and domo_group.id,
            permission=permission,
            auth=self.auth,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
            parent_class=self.__class__.__name__,
        )

    async def query_documents(
        self,
        query: dict = None,
        return_raw: bool = False,
        try_auto_share=False,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient = None,
    ):
        documents = []
        loop_retry = 0
        while loop_retry <= 1 and not documents:
            try:
                res = await appdb_routes.get_documents_from_collection(
                    auth=self.auth,
                    collection_id=self.id,
                    query=query,
                    debug_api=debug_api,
                    debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                    session=session,
                )

                documents = res.response

            except appdb_routes.AppDb_GET_Exception as e:
                if try_auto_share:
                    await self.share_collection(debug_api=debug_api)
                    await asyncio.sleep(2)

                loop_retry += 1

                if loop_retry > 1:
                    raise e

            if return_raw:
                return res

            self.domo_documents = await dmce.gather_with_concurrency(
                *[
                    AppDbDocument.get_by_id(
                        collection_id=self.id, document_id=doc["id"], auth=self.auth
                    )
                    for doc in documents
                ],
                n=60,
            )

            return self.domo_documents


async def share_collection(
    self,
    domo_user=None,
    domo_group=None,
    permission: appdb_routes.Collection_Permission_Enum = appdb_routes.Collection_Permission_Enum.READ_CONTENT,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    session: httpx.AsyncClient = None,
):
    return await appdb_routes.modify_collection_permissions(
        collection_id=self.id,
        user_id=(domo_user and domo_user.id)
        or (await self.auth.who_am_i()).response["id"],
        group_id=domo_group and domo_group.id,
        permission=permission,
        auth=self.auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=self.__class__.__name__,
    )


async def query_documents(
    self,
    query: dict = None,
    return_raw: bool = False,
    try_auto_share=False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    session: httpx.AsyncClient = None,
):
    documents = []
    loop_retry = 0
    while loop_retry <= 1 and not documents:
        try:
            res = await appdb_routes.get_documents_from_collection(
                auth=self.auth,
                collection_id=self.id,
                query=query,
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                session=session,
            )

            documents = res.response

        except appdb_routes.AppDb_GET_Exception as e:
            if try_auto_share:
                await self.share_collection(debug_api=debug_api)
                await asyncio.sleep(2)

            loop_retry += 1

            if loop_retry > 1:
                raise e

        if return_raw:
            return res

        self.domo_documents = await dmce.gather_with_concurrency(
            *[
                AppDbDocument.get_by_id(
                    collection_id=self.id, document_id=doc["id"], auth=self.auth
                )
                for doc in documents
            ],
            n=60,
        )

        return self.domo_documents


async def upsert(
    cls,
    auth: DomoAuth,
    collection_id,
    content: dict,
    identity_columns: list[str],
    session: httpx.AsyncClient = None,
    debug_api=False,
    debug_num_stacks_to_drop=3,
    return_raw: bool = False,
):
    domo_doc = None

    query = {f"content.{col}": content[col] for col in identity_columns}

    domo_collection = await AppDbCollection.get_by_id(
        auth=auth, collection_id=collection_id, return_raw=False
    )

    res = await domo_collection.query_documents(query=query, debug_api=debug_api)
    domo_doc = res[0] if res and len(res) > 0 else None

    if not domo_doc:
        return await cls.create_document(
            content=content,
            collection_id=collection_id,
            auth=auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=return_raw,
        )

    return await domo_doc.update_document(
        session=session, content=content, debug_api=debug_api, return_raw=return_raw
    )


@dataclass
class AppDbCollections:
    @classmethod
    async def get_collections(
        cls,
        auth: DomoAuth,
        datastore_id: str | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        debug_num_stacks_to_drop=1,
        return_raw: bool = False,
    ):
        res = await appdb_routes.get_collections(
            auth=auth,
            datastore_id=datastore_id,
            parent_class=cls.__class__.__name__,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return await dmce.gather_with_concurrency(
            *[
                AppDbCollection.get_by_id(collection_id=obj["id"], auth=auth)
                for obj in res.response
            ],
            n=10,
        )



__all__ = ["DomoGrant", "DomoGrants"]

from dataclasses import dataclass, field

import httpx

from domolibrary2.base.entities import DomoBase, DomoManager

from ...auth import DomoAuth as dmda
from ...routes import grant as grant_routes
from ...utils import DictDot as util_dd


@dataclass
class DomoGrant(DomoBase):
    id: str
    display_group: str | None = None
    title: str | None = None
    depends_on_ls: list[str] | None = None
    description: str | None = None
    role_membership_ls: list[str] | None = field(default=None)

    def __post_init__(self):
        self.id = str(self.id)

    def __eq__(self, other):
        if not isinstance(other, DomoGrant):
            return False

        return self.id == other.id

    @classmethod
    def from_dict(cls, obj):
        dd = obj
        if not isinstance(dd, util_dd.DictDot):
            dd = util_dd.DictDot(obj)

        return cls(
            id=dd.authority,
            display_group=dd.authorityUIGroup,
            depends_on_ls=dd.dependsOnAuthorities,
            title=dd.title,
            description=dd.description,
            role_membership_ls=[str(role) for role in dd.roleIds],
        )


@dataclass
class DomoGrants(DomoManager):
    auth: dmda = field(repr=False)

    grants: list[DomoGrant] | None = field(default=None)

    async def get(
        self,
        session: httpx.AsyncClient | None = None,
        debug_api: bool = False,
        return_raw: bool = False,
    ):
        res = await grant_routes.get_grants(
            auth=self.auth, debug_api=debug_api, session=session
        )

        if return_raw or not res.is_success:
            return res

        self.grants = [DomoGrant.from_dict(obj) for obj in res.response]

        return self.grants



__all__ = [
    "validate_ip_or_cidr",
    "DomoAllowlist",
]

import ipaddress
from dataclasses import dataclass, field

import httpx

from ...auth import DomoAuth
from ...routes.instance_config import allowlist as allowlist_routes


def validate_ip_or_cidr(ip: str):
    try:
        # Try IPv4 address
        ipaddress.IPv4Address(ip)
        return True
    except ValueError:
        try:
            # Try IPv4 network (CIDR)
            ipaddress.IPv4Network(ip, strict=False)
            return True
        except ValueError as e:
            raise ValueError(f"Invalid IP/CIDR entry: {ip}") from e


@dataclass
class DomoAllowlist:
    """A class for managing the Domo instance IP allowlist configuration.

    This class provides methods to get, set, add, and remove IP addresses/CIDR ranges
    from the instance allowlist, as well as manage the filter all traffic setting.

    Note: Unlike most Domo entities, the allowlist is a singleton configuration per instance
    and does not have an 'id' field.
    """

    auth: DomoAuth = field(repr=False)
    allowlist: list[str] = field(default_factory=list)
    is_filter_all_traffic_enabled: bool | None = None
    raw: dict = field(default_factory=dict, repr=False)

    @property
    def display_url(self) -> str:
        """Return the URL to the allowlist configuration page in Domo."""
        return f"https://{self.auth.domo_instance}.domo.com/admin/security/settings"

    @classmethod
    def from_dict(cls, auth: DomoAuth, obj: dict) -> "DomoAllowlist":
        """Create a DomoAllowlist instance from a dictionary representation.

        Args:
            auth: Authentication object for API requests
            obj: Dictionary containing allowlist data

        Returns:
            DomoAllowlist instance
        """
        return cls(
            auth=auth,
            allowlist=obj.get("allowlist") or obj.get("addresses", []),
            is_filter_all_traffic_enabled=obj.get("is_filter_all_traffic_enabled"),
            raw=obj,
        )

    async def get(
        self,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient | None = None,
    ) -> list[str]:
        """
        retrieves the allowlist for an instance
        """

        res = await allowlist_routes.get_allowlist(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            return_raw=return_raw,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.allowlist = [str(s) for s in res.response]

        return self.allowlist

    async def set(
        self,
        ip_address_ls: list[str],
        is_suppress_errors: bool = False,
        debug_api: bool = False,
        debug_prn: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient | None = None,
    ):
        for ip in ip_address_ls:
            try:
                validate_ip_or_cidr(ip)
            except ValueError as ve:
                if not is_suppress_errors:
                    raise ve

                print(f"skipping invalid entry: {ip}")
                ip_address_ls.remove(ip)
                continue

        # get current allowlist
        await self.get(session=session)

        if sorted(ip_address_ls) == sorted(self.allowlist):
            if debug_prn:
                print("no changes to allowlist detected, skipping update")
            return self.allowlist

        await allowlist_routes.set_allowlist(
            ip_address_ls=ip_address_ls,
            auth=self.auth,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

        return await self.get(
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

    async def add_ips(
        self,
        ip_address_ls: str,
        is_suppress_errors: bool = False,
        debug_api: bool = False,
        debug_prn: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> list[str]:
        """
        adds an IP or CIDR to the allowlist
        """

        new_allowlist = await self.get()

        for ip in ip_address_ls:
            if ip in new_allowlist:
                continue

            new_allowlist.append(ip)

        return await self.set(
            ip_address_ls=new_allowlist,
            session=session,
            debug_api=debug_api,
            debug_prn=debug_prn,
            is_suppress_errors=is_suppress_errors,
        )

    async def remove_ips(
        self,
        ip_address_ls: str,
        debug_prn: bool = False,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        is_suppress_errors: bool = False,
    ) -> list[str]:
        """
        removes an IP or CIDR to the allowlist
        """

        allowlist = await self.get()

        for ip in ip_address_ls:
            if ip not in allowlist:
                continue

            allowlist.remove(ip)

        return await self.set(
            ip_address_ls=allowlist,
            debug_api=debug_api,
            session=session,
            is_suppress_errors=is_suppress_errors,
            debug_prn=debug_prn,
        )

    async def get_is_filter_all_traffic_enabled(
        self,
        debug_api: bool = False,
        return_raw: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> bool:
        """
        retrieves whether the "filter all traffic" setting is enabled
        """

        res = await allowlist_routes.get_allowlist_is_filter_all_traffic_enabled(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            return_raw=return_raw,
            parent_class=self.__class__.__name__,
        )

        if return_raw:
            return res

        self.is_filter_all_traffic_enabled = bool(res.response["is_enabled"])

        return self.is_filter_all_traffic_enabled

    async def toggle_is_filter_all_traffic_enabled(
        self,
        is_enabled: bool,
        debug_api: bool = False,
        debug_prn: bool = False,
        return_raw: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> bool:
        """
        retrieves whether the "filter all traffic" setting is enabled
        """

        await self.get_is_filter_all_traffic_enabled(session=session)

        if self.is_filter_all_traffic_enabled == is_enabled:
            if debug_prn:
                print("no action required")

            return bool(self.is_filter_all_traffic_enabled)

        res = await allowlist_routes.toggle_allowlist_is_filter_all_traffic_enabled(
            auth=self.auth,
            is_enabled=is_enabled,
            debug_api=debug_api,
            session=session,
            return_raw=return_raw,
            parent_class=self.__class__.__name__,
        )

        if return_raw:
            return res

        self.is_filter_all_traffic_enabled = res.response["is_enabled"]

        return bool(self.is_filter_all_traffic_enabled)



"""Page classes and functionality.

This module provides comprehensive page management functionality for Domo pages,
organized into focused submodules:

- exceptions: Page-related exception classes
- core: Main DomoPage entity class with basic operations
- pages: DomoPages collection and hierarchy operations
- access: Access control and sharing functionality
- content: Content management and data operations

Classes:
    DomoPage: Main page entity class
    DomoPages: Collection class for managing multiple pages
    DomoPage_GetRecursive: Exception for recursive operation conflicts
    Page_NoAccess: Exception for page access denial

Example:
    Basic page usage:

        >>> from domolibrary2.classes.DomoPage import DomoPage
        >>> page = await DomoPage.get_by_id(page_id="123", auth=auth)
        >>> print(page.display_url())

    Managing page collections:

        >>> from domolibrary2.classes.DomoPage import DomoPages
        >>> pages = await DomoPages(auth=auth).get()
        >>> print(f"Found {len(pages)} pages")

    Access control:

        >>> access_info = await page.get_accesslist()
        >>> await page.share(domo_users=[user1, user2])
"""

__all__ = [
    "DomoPage_GetRecursive",
    "DomoPage",
    "DomoPages",
    "Page_NoAccess",
    "access",
    "content",
]

# Import and attach functionality modules
from . import access, content

# Import core classes
from .core import DomoPage

# Import exceptions
from .exceptions import DomoPage_GetRecursive, Page_NoAccess
from .pages import DomoPages

# Attach methods to DomoPage class
DomoPage.test_page_access = access.test_page_access
DomoPage.get_accesslist = access.get_accesslist
DomoPage.share = access.share

DomoPage.get_cards = content.get_cards
DomoPage.get_datasets = content.get_datasets
DomoPage.update_layout = content.update_layout
DomoPage.add_owner = content.add_owner



"""
Page Route Functions

This module provides functions for managing Domo pages including retrieval,
layout management, access control, and sharing operations.

The module is organized into functional areas:
- core: Basic page retrieval and definition functions
- access: Access control and permissions functions
- crud: Layout management, locks, and ownership functions
- exceptions: All page-related exception classes

Functions:
    get_pages_adminsummary: Retrieve all pages visible to the user
    get_page_by_id: Retrieve a specific page by ID
    get_page_definition: Retrieve detailed page definition
    get_page_access_test: Test page access permissions
    get_page_access_list: Retrieve page access list with users and groups
    update_page_layout: Update page layout configuration
    put_writelock: Set write lock on a page layout
    delete_writelock: Remove write lock from a page layout
    add_page_owner: Add owners to pages

Exception Classes:
    Page_GET_Error: Raised when page retrieval fails
    SearchPageNotFoundError: Raised when page search returns no results
    Page_CRUD_Error: Raised when page create/update/delete operations fail
    PageSharing_Error: Raised when page sharing operations fail
"""

from .access import (
    add_page_owner,
    get_page_access_list,
    get_page_access_test,
)
from .core import (
    get_page_by_id,
    get_page_definition,
    get_pages_adminsummary,
)
from .crud import (
    delete_writelock,
    put_writelock,
    update_page_layout,
)
from .exceptions import (
    Page_CRUD_Error,
    Page_GET_Error,
    PageSharing_Error,
    SearchPageNotFoundError,
)

__all__ = [
    "get_page_access_test",
    "get_page_access_list",
    "add_page_owner",
    "get_pages_adminsummary",
    "get_page_by_id",
    "get_page_definition",
    "update_page_layout",
    "put_writelock",
    "delete_writelock",
    "Page_GET_Error",
    "SearchPageNotFoundError",
    "Page_CRUD_Error",
    "PageSharing_Error",
]



"""Base classes and utilities for stream configuration mappings.

This module contains:
- StreamConfig_Mapping: Base class for all provider-specific mappings (OLD - deprecated)
- StreamConfig_Base: New base class for typed stream configs (NEW - follows AccountConfig pattern)
- StreamConfig: Individual stream configuration item (OLD - for list-based configs)
- StreamConfig_Mappings: Enum for accessing registered mappings
- register_mapping: Decorator for registering new mappings
- register_stream_config: Decorator for registering new typed config classes (NEW)
"""

from __future__ import annotations

import re
from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Any

from sqlglot import exp, parse_one

from ....base.base import DomoBase, DomoEnumMixin
from ....routes.stream import Stream_CRUD_Error, Stream_GET_Error

__all__ = [
    # New pattern (AccountConfig-style)
    "StreamConfig_Base",
    "register_stream_config",
    # Old pattern (deprecated but kept for compatibility)
    "StreamConfig_Mapping",
    "StreamConfig_Mappings",
    "StreamConfig",
    "register_mapping",
    # Route exceptions
    "Stream_GET_Error",
    "Stream_CRUD_Error",
]

# ============================================================================
# OLD PATTERN: Mapping Registry (deprecated but kept for compatibility)
# ============================================================================

# Registry to store mapping classes (OLD)
_MAPPING_REGISTRY: dict[str, type[StreamConfig_Mapping]] = {}


def register_mapping(data_provider_type: str):
    """Decorator to register a StreamConfig_Mapping subclass (OLD pattern).

    DEPRECATED: Use register_stream_config() for new config classes.

    Args:
        data_provider_type: The data provider type identifier (e.g., 'snowflake')

    Example:
        @register_mapping('snowflake')
        @dataclass
        class SnowflakeMapping(StreamConfig_Mapping):
            sql: str = "query"
            warehouse: str = "warehouseName"
            database_name: str = "databaseName"
    """

    def decorator(cls: type[StreamConfig_Mapping]) -> type[StreamConfig_Mapping]:
        _MAPPING_REGISTRY[data_provider_type] = cls
        return cls

    return decorator


# ============================================================================
# NEW PATTERN: Config Registry (follows AccountConfig pattern)
# ============================================================================

# Registry to store typed config classes (NEW)
_CONFIG_REGISTRY: dict[str, type[StreamConfig_Base]] = {}


def register_stream_config(data_provider_type: str):
    """Decorator to register a StreamConfig_Base subclass (NEW pattern).

    Args:
        data_provider_type: The data provider type identifier (e.g., 'snowflake')

    Example:
        @register_stream_config('snowflake')
        @dataclass
        class Snowflake_StreamConfig(StreamConfig_Base):
            data_provider_type: str = "snowflake"
            query: str = None
            database_name: str = None
    """

    def decorator(cls: type[StreamConfig_Base]) -> type[StreamConfig_Base]:
        _CONFIG_REGISTRY[data_provider_type] = cls
        return cls

    return decorator


# ============================================================================
# Utility Functions (for new pattern)
# ============================================================================


def _camel_to_snake(name: str) -> str:
    """Convert camelCase to snake_case.

    Args:
        name: camelCase string

    Returns:
        snake_case string

    Example:
        >>> _camel_to_snake("databaseName")
        "database_name"
        >>> _camel_to_snake("warehouseName")
        "warehouse_name"
    """
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def _snake_to_camel(name: str) -> str:
    """Convert snake_case to camelCase.

    Args:
        name: snake_case string

    Returns:
        camelCase string

    Example:
        >>> _snake_to_camel("database_name")
        "databaseName"
        >>> _snake_to_camel("warehouse_name")
        "warehouseName"
    """
    components = name.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


# ============================================================================
# NEW: StreamConfig_Base (follows AccountConfig pattern)
# ============================================================================


@dataclass
class StreamConfig_Base(DomoBase):
    """Base class for typed stream configurations (NEW pattern).

    Similar to DomoAccount_Config, this provides a consistent interface
    for handling stream execution parameters across different data providers.

    This is the NEW pattern that provides:
    - Type-safe attribute access (stream_config.query)
    - IDE autocomplete
    - Single config object instead of list of configs
    - Validation and computed properties

    Example:
        @register_stream_config("snowflake")
        @dataclass
        class Snowflake_StreamConfig(StreamConfig_Base):
            data_provider_type: str = "snowflake"
            query: str = None
            database_name: str = None
            warehouse: str = None
    """

    data_provider_type: str = None
    parent: Any = field(repr=False, default=None)
    raw: dict = field(default_factory=dict, repr=False)

    # Internal fields
    _field_map: dict = field(default_factory=dict, repr=False, init=False)
    _fields_for_export: list[str] = field(default_factory=list, repr=False, init=False)

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None):
        """Create StreamConfig from dictionary of config parameters.

        Args:
            obj: Dict with config parameter names as keys (e.g., {"query": "SELECT...", "databaseName": "mydb"})
            parent: Parent stream object

        Returns:
            StreamConfig instance with typed attributes

        Example:
            >>> config_dict = {"query": "SELECT *", "databaseName": "SA_PRD"}
            >>> config = Snowflake_StreamConfig.from_dict(config_dict)
            >>> config.query  # "SELECT *"
            >>> config.database_name  # "SA_PRD"
        """
        # Get field map from class definition
        field_map = {}
        if "_field_map" in cls.__dataclass_fields__:
            field_map_field = cls.__dataclass_fields__["_field_map"]
            if (
                hasattr(field_map_field, "default_factory")
                and field_map_field.default_factory
            ):
                field_map = field_map_field.default_factory()

        # Convert keys using field_map or auto-convert camelCase to snake_case
        init_kwargs = {}
        for api_key, api_value in obj.items():
            # Check if there's a custom mapping
            python_attr = field_map.get(api_key)
            if not python_attr:
                # Auto-convert camelCase to snake_case
                python_attr = _camel_to_snake(api_key)

            # Only set if it's a field on the class
            if python_attr in cls.__dataclass_fields__:
                init_kwargs[python_attr] = api_value

        return cls(parent=parent, raw=obj, **init_kwargs)

    def to_dict(self) -> list[dict]:
        """Convert to list of config dictionaries for API submission.

        Returns:
            List of dicts with 'name', 'type', 'value' keys

        Example:
            >>> config = Snowflake_StreamConfig(query="SELECT *", database_name="SA_PRD")
            >>> config.to_dict()
            [{"name": "query", "type": "string", "value": "SELECT *"},
             {"name": "databaseName", "type": "string", "value": "SA_PRD"}]
        """
        result = []

        # Get reverse field map (python_attr -> api_key)
        reverse_map = {v: k for k, v in self._field_map.items()}

        # Get fields to export
        export_fields = self._fields_for_export or [
            f
            for f in self.__dataclass_fields__.keys()
            if not f.startswith("_")
            and f not in ["data_provider_type", "parent", "raw"]
        ]

        for attr_name in export_fields:
            value = getattr(self, attr_name, None)
            if value is not None:
                # Get API key name (reverse mapping or convert snake_case to camelCase)
                api_key = reverse_map.get(attr_name, _snake_to_camel(attr_name))
                result.append({"name": api_key, "type": "string", "value": str(value)})

        return result


# ============================================================================
# OLD: StreamConfig_Mapping (deprecated but kept for compatibility)
# ============================================================================


@dataclass
class StreamConfig_Mapping(DomoBase):
    """Base class for stream configuration mappings.

    Maps Python attribute names to stream configuration parameter names.
    Subclass this for each data provider type.
    """

    data_provider_type: str | None = None

    sql: str = None
    warehouse: str = None
    database_name: str = None
    s3_bucket_category: str = None

    is_default: bool = False

    table_name: str = None
    src_url: str = None
    google_sheets_file_name: str = None
    adobe_report_suite_id: str = None
    qualtrics_survey_id: str = None

    def search_keys_by_value(
        self,
        value_to_search: str,
    ) -> StreamConfig_Mapping | None:
        """Search for Python attribute name by stream config parameter name.

        Args:
            value_to_search: Stream config parameter name to search for

        Returns:
            Python attribute name if found, None otherwise
        """
        if self.is_default:
            if value_to_search in ["enteredCustomQuery", "query", "customQuery"]:
                return "sql"

        return next(
            (key for key, value in asdict(self).items() if value == value_to_search),
            None,
        )


# ============================================================================
# StreamConfig and StreamConfig_Mappings must be defined before platform imports
# ============================================================================


@dataclass
class StreamConfig:
    """Individual stream configuration parameter.

    Represents a single configuration parameter in a stream execution,
    such as a SQL query, database name, warehouse, etc.
    """

    stream_category: str
    name: str
    type: str
    value: str
    value_clean: str = None
    parent: Any = field(repr=False, default=None)

    def __post_init__(self):
        # self.value_clean = self.value.replace("\n", " ")
        # sc.value_clean = re.sub(" +", " ", sc.value_clean)

        if self.stream_category == "sql" and self.parent:
            self.process_sql()

    def process_sql(self):
        """Extract table names from SQL query using sqlglot parser."""
        if not self.parent:
            return None

        self.parent.configuration_query = self.value

        try:
            for table in parse_one(self.value).find_all(exp.Table):
                self.parent.configuration_tables.append(table.name.lower())
                self.parent.configuration_tables = sorted(
                    list(set(self.parent.configuration_tables))
                )
        except Exception:
            return None

        return self.parent.configuration_tables

    @classmethod
    def from_json(cls, obj: dict, data_provider_type: str, parent_stream: Any = None):
        """Create StreamConfig from API JSON response.
        Args:
            obj: JSON object with 'name', 'type', 'value' keys
            data_provider_type: Data provider type (e.g., 'snowflake')
            parent_stream: Parent stream object
        Returns:
            StreamConfig instance
        """
        config_name = obj["name"]

        # Use standard enum access - triggers _missing_ if not found
        mapping_enum = StreamConfig_Mappings.get(data_provider_type)

        stream_category = "default"
        if mapping_enum and mapping_enum.value:
            stream_category = mapping_enum.value.search_keys_by_value(config_name)

            if parent_stream:
                parent_stream.has_mapping = True

        return cls(
            stream_category=stream_category,
            name=config_name,
            type=obj["type"],
            value=obj["value"],
            parent=parent_stream,
        )

    def to_dict(self):
        """Convert to dictionary for API submission.
        Returns:
            Dict with 'field', 'key', 'value' keys
        """
        return {"field": self.stream_category, "key": self.name, "value": self.value}


# ============================================================================
# Import platform-specific mappings to trigger registration
# ============================================================================

# Import all mappings from the stream_configs submodules
# This triggers the @register_mapping decorators and populates _MAPPING_REGISTRY
from . import _default, aws, domo, google, other, snowflake  # noqa: E402, F401

# ============================================================================
# StreamConfig_Mappings Enum (Auto-generated from registry)
# ============================================================================


class StreamConfig_Mappings(DomoEnumMixin, Enum):
    """Enum of all registered stream config mappings.

    This enum is automatically populated from the registry created by
    @register_mapping decorators. To add a new mapping, simply create a
    new subclass with the @register_mapping decorator in the appropriate
    platform file.
    Usage:
        mapping = StreamConfig_Mappings.get('snowflake')
        sql_param = mapping.value.sql  # "query"
    """

    # Explicit default member to prevent AttributeError
    default = None  # Will be set dynamically via _missing_

    @classmethod
    def _missing_(cls, value):
        """Handle missing enum values by searching the registry."""
        alt_search = value.lower().replace("-", "_")

        # Try direct registry lookup
        if value in _MAPPING_REGISTRY:
            mapping_cls = _MAPPING_REGISTRY[value]
            return cls._create_pseudo_member(value, mapping_cls())

        # Try normalized search
        for key, mapping_cls in _MAPPING_REGISTRY.items():
            if key.lower().replace("-", "_") == alt_search:
                return cls._create_pseudo_member(key, mapping_cls())

        # Return default
        return cls.default



"""
File and Folder Operation Utilities

This module provides utilities for file and folder operations including creation,
manipulation, and archive handling. All functions include proper error handling
and support for various file operations.

Functions:
    upsert_folder: Create folder if it doesn't exist, with optional replacement
    upsert_file: Create or update file with content
    change_extension: Change file extension
    export_zip_binary_contents: Save binary content as zip file
    download_zip: Extract or save zip file contents

Exception Classes:
    FileOperationError: Raised when file operations fail

Example:
    >>> # Create folder and file
    >>> upsert_folder("/path/to/folder")
    >>> upsert_file("/path/to/file.txt", "Hello, World!")

    >>> # Change file extension
    >>> new_path = change_extension("/path/to/file.txt", ".md")
    >>> print(new_path)  # "/path/to/file.md"

    >>> # Handle zip files
    >>> files = download_zip("/output/folder", zip_bytes_content=zip_data)
"""

__all__ = [
    "upsert_folder",
    "upsert_file",
    "change_extension",
    "export_zip_binary_contents",
    "download_zip",
]


import io
import os
import pathlib
import shutil
import zipfile
from typing import Callable

from .exceptions import FileOperationError


def upsert_folder(
    folder_path: str, debug_prn: bool = False, replace_folder: bool = False
) -> None:
    """
    Create folder if it doesn't exist, with optional replacement of existing folder.

    Args:
        folder_path (str): Path to folder to create or ensure exists
        debug_prn (bool): Print debug information if True
        replace_folder (bool): Remove and recreate folder if it exists

    Raises:
        FileOperationError: If folder creation fails

    Example:
        >>> upsert_folder("/path/to/new/folder")
        >>> upsert_folder("/path/to/folder", replace_folder=True)  # Recreates folder

    Note:
        If folder_path contains a filename, the directory portion is extracted.
        Creates all necessary parent directories.
    """
    try:
        # Extract directory path if a filename is included
        folder_path = os.path.dirname(folder_path)

        if (
            replace_folder
            and os.path.exists(folder_path)
            and os.path.isdir(folder_path)
        ):
            folder_path = os.path.join(folder_path, "")
            shutil.rmtree(folder_path)

        if debug_prn:
            print(
                {
                    "upsert_folder": os.path.abspath(folder_path),
                    "is_exist": os.path.exists(folder_path),
                }
            )

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

    except (OSError, PermissionError) as e:
        raise FileOperationError("create folder", folder_path, str(e))


def upsert_file(
    file_path: str,
    content: object = None,
    write_fn: Callable = None,
    file_update_method: str = "w",
    encoding: str = "utf-8",
    debug_prn: bool = False,
    replace_folder: bool = False,
) -> bool:
    """
    Create or update file with specified content.

    Args:
        file_path (str): Path to file to create or update
        content (Any, optional): Content to write to file
        write_fn (Callable, optional): Custom function for writing content
        file_update_method (str): File open mode ('w', 'a', 'wb', etc.)
        encoding (str): Text encoding for file operations
        debug_prn (bool): Print debug information if True
        replace_folder (bool): Replace parent folder if it exists

    Returns:
        bool: True if operation succeeded

    Raises:
        FileOperationError: If file creation or writing fails

    Example:
        >>> # Simple file creation
        >>> upsert_file("/path/to/file.txt", "Hello, World!")
        True

        >>> # Custom write function
        >>> def custom_writer(file, content):
        ...     file.write(f"Custom: {content}")
        ...     return True
        >>> upsert_file("/path/to/file.txt", "data", write_fn=custom_writer)
        True

        >>> # Binary file
        >>> upsert_file("/path/to/file.bin", b"binary data", file_update_method="wb")
        True
    """
    try:
        upsert_folder(
            folder_path=file_path, debug_prn=debug_prn, replace_folder=replace_folder
        )

        with open(file_path, file_update_method, encoding=encoding) as file:
            if not write_fn:
                file.write(content or "")
                return True
            return write_fn(file=file, content=content)

    except (OSError, PermissionError, UnicodeError) as e:
        raise FileOperationError("create or write file", file_path, str(e))


def change_extension(file_path: str, new_extension: str) -> str:
    """
    Change the extension of a file path.

    Args:
        file_path (str): Original file path
        new_extension (str): New extension (with or without leading dot)

    Returns:
        str: File path with new extension

    Example:
        >>> change_extension("/path/to/file.txt", ".md")
        '/path/to/file.md'
        >>> change_extension("/path/to/file.txt", "json")
        '/path/to/file.json'
    """
    path = pathlib.PurePath(file_path)

    # Ensure extension starts with dot
    new_extension = (
        new_extension if new_extension.startswith(".") else "." + new_extension
    )

    new_file_path = path.with_suffix(new_extension)
    return str(new_file_path)


def export_zip_binary_contents(output_folder: str, zip_bytes_content: bytes) -> str:
    """
    Save binary zip content to a file.

    Args:
        output_folder (str): Path where zip file should be saved
        zip_bytes_content (bytes): Binary zip file content

    Returns:
        str: Success message with file location

    Raises:
        FileOperationError: If zip file creation fails

    Example:
        >>> with open("source.zip", "rb") as f:
        ...     zip_data = f.read()
        >>> result = export_zip_binary_contents("/output/archive", zip_data)
        >>> print(result)  # "successfully downloaded to /output/archive.zip"
    """
    try:
        output_folder = change_extension(output_folder, ".zip")
        upsert_folder(output_folder)

        with open(output_folder, "wb+") as f:
            f.write(zip_bytes_content)

        return f"successfully downloaded to {output_folder}"

    except (OSError, PermissionError) as e:
        raise FileOperationError("export zip content", output_folder, str(e))


def download_zip(
    output_folder: str,
    zip_bytes_content: bytes | None = None,
    existing_zip_file_path: str | None = None,
    is_unpack_archive: bool = True,
) -> str | list[str]:
    """
    Save bytes content to a zip file and optionally extract contents.

    Args:
        output_folder (str): Folder where files should be saved/extracted
        zip_bytes_content (bytes, optional): Zip file as bytes
        existing_zip_file_path (str, optional): Path to existing zip file
        is_unpack_archive (bool): Whether to extract the archive contents

    Returns:
        Union[str, list[str]]: Success message or list of extracted files

    Raises:
        FileOperationError: If zip operations fail
        ValueError: If neither zip_bytes_content nor existing_zip_file_path provided

    Example:
        >>> # Extract zip from bytes
        >>> files = download_zip("/output", zip_bytes_content=zip_data)
        >>> print(files)  # ['file1.txt', 'file2.txt', 'folder/']

        >>> # Just save zip file without extracting
        >>> result = download_zip("/output", zip_data, is_unpack_archive=False)
        >>> print(result)  # "successfully downloaded to /output.zip"

        >>> # Extract existing zip file
        >>> files = download_zip("/output", existing_zip_file_path="/path/to/archive.zip")
    """
    try:
        # Ensure output folder ends with slash for consistency
        output_folder = (
            output_folder if output_folder.endswith("/") else output_folder + "/"
        )

        # Handle case where we just want to save the zip file
        if not is_unpack_archive:
            if zip_bytes_content:
                return export_zip_binary_contents(
                    output_folder=output_folder, zip_bytes_content=zip_bytes_content
                )

            if existing_zip_file_path and os.path.exists(existing_zip_file_path):
                output_zip = os.path.join(output_folder, "archive.zip")
                upsert_folder(output_folder)
                shutil.copy(src=existing_zip_file_path, dst=output_zip)
                return f"zip available at {output_zip}"

        # Extract zip contents
        zip_file = None
        upsert_folder(output_folder)

        if zip_bytes_content:
            zip_file = zipfile.ZipFile(io.BytesIO(zip_bytes_content), "r")
        elif existing_zip_file_path and os.path.exists(existing_zip_file_path):
            zip_file = zipfile.ZipFile(existing_zip_file_path, "r")

        if not zip_file:
            raise ValueError(
                "Must provide either zip_bytes_content or existing_zip_file_path"
            )

        # Extract all files
        zip_file.extractall(output_folder)
        file_list = os.listdir(output_folder)
        zip_file.close()

        return file_list

    except (OSError, PermissionError, zipfile.BadZipFile) as e:
        raise FileOperationError("download and extract zip", output_folder, str(e))
    except ValueError as e:
        raise e  # Re-raise ValueError as-is



"""DomoInstanceConfig classes for managing instance-level configuration."""

from . import (
    access_token,
    allowlist,
    api_client,
    bootstrap,
    core,
    instance_switcher,
    mfa,
    role,
    role_grant,
    scheduler_policies,
    sso,
    toggle,
    user_attributes,
)

# Import main class
from .core import DomoInstanceConfig

__all__ = [
    "DomoInstanceConfig",
    "access_token",
    "allowlist",
    "api_client",
    "bootstrap",
    "core",
    "instance_switcher",
    "mfa",
    "role",
    "role_grant",
    "scheduler_policies",
    "sso",
    "toggle",
    "user_attributes",
]



# agent_models.py - Pydantic models for agent outputs

from typing import Optional

from pydantic import BaseModel, Field


class ConversionPlan(BaseModel):
    """Output model for Orchestrator Agent."""

    collection_name: str
    total_requests: int
    processing_strategy: str = Field(description="sequential or parallel")
    estimated_complexity: str = Field(description="low, medium, high")
    next_step: str


class ParsedCollection(BaseModel):
    """Output model for Parser Agent."""

    collection: dict = Field(description="Parsed PostmanCollection as dict")
    total_requests: int
    folder_structure: list[str] = Field(description="List of folder paths")
    parsing_issues: list[str] = Field(default_factory=list)
    collection_variables: dict[str, str]
    collection_auth: Optional[dict] = None


class ValidationReport(BaseModel):
    """Output model for Validation Agent."""

    is_valid: bool
    errors: list[str] = Field(default_factory=list)
    warnings: list[str] = Field(default_factory=list)
    recommendations: list[str] = Field(default_factory=list)
    passed_checks: list[str] = Field(default_factory=list)


class StructureAnalysis(BaseModel):
    """Output model for Structure Analyzer Agent."""

    folder_hierarchy: dict[str, list[str]]
    request_groupings: dict[str, list[str]]
    naming_patterns: list[str] = Field(description="Identified naming conventions")
    suggested_module_structure: dict[str, list[str]]
    complexity_score: int = Field(ge=1, le=10)


class AuthAnalysis(BaseModel):
    """Output model for Authentication Analyzer Agent."""

    auth_types: list[str] = Field(
        description="Types of auth used: bearer, basic, apikey, etc."
    )
    auth_locations: list[str] = Field(description="collection, folder, request")
    requires_token_refresh: bool
    auth_variables: list[str] = Field(description="Variables used in auth")
    suggested_auth_class: str = Field(description="Recommended auth class name")
    auth_implementation_notes: str


class ParameterAnalysis(BaseModel):
    """Output model for Parameter Analyzer Agent."""

    common_params: dict[str, list[str]] = Field(
        description="Parameters used across multiple requests"
    )
    required_params: dict[str, list[str]] = Field(
        description="Request name to required params"
    )
    optional_params: dict[str, list[str]] = Field(
        description="Request name to optional params"
    )
    param_types: dict[str, str] = Field(description="Inferred parameter types")
    default_values: dict[str, str] = Field(description="Default values for params")
    path_variables: dict[str, list[str]] = Field(
        description="Path variables per request"
    )


class HeaderAnalysis(BaseModel):
    """Output model for Header Analyzer Agent."""

    common_headers: dict[str, list[str]] = Field(
        description="Headers used across requests"
    )
    required_headers: list[str] = Field(description="Headers needed for all requests")
    optional_headers: dict[str, list[str]] = Field(
        description="Request-specific headers"
    )
    header_variables: list[str] = Field(description="Variables in header values")
    content_types: list[str] = Field(description="Content-Type values found")
    auth_headers: list[str] = Field(description="Headers used for authentication")


class AggregatedAnalysis(BaseModel):
    """Output model for Analysis Aggregator Agent."""

    structure: dict = Field(description="From Structure Analyzer")
    authentication: dict = Field(description="From Auth Analyzer")
    parameters: dict = Field(description="From Parameter Analyzer")
    headers: dict = Field(description="From Header Analyzer")

    code_generation_strategy: str = Field(description="Recommended approach")
    module_organization: dict[str, list[str]]
    shared_components: list[str] = Field(description="Components to generate once")
    per_request_customization: dict[str, dict]
    complexity_assessment: str


class GeneratedCode(BaseModel):
    """Output model for Code Generator Agent."""

    function_name: str
    function_code: str = Field(description="Complete Python function code")
    imports_needed: list[str] = Field(description="Import statements required")
    dependencies: list[str] = Field(description="Other functions this depends on")
    docstring: str
    type_hints: dict[str, str] = Field(description="Parameter name to type hint")
    complexity: str = Field(description="low, medium, high")


class GeneratedTests(BaseModel):
    """Output model for Test Generator Agent."""

    test_function_name: str
    test_code: str = Field(description="Complete test function code")
    test_fixtures: list[str] = Field(description="Fixtures needed")
    mock_data: dict = Field(description="Mock data for testing")
    assertions: list[str] = Field(description="Key assertions made")
    test_coverage_areas: list[str] = Field(description="What aspects are tested")


class CodeValidationResult(BaseModel):
    """Output model for Code Validator Agent."""

    is_valid: bool
    syntax_errors: list[str] = Field(default_factory=list)
    type_errors: list[str] = Field(default_factory=list)
    style_issues: list[str] = Field(default_factory=list)
    security_concerns: list[str] = Field(default_factory=list)
    suggestions: list[str] = Field(default_factory=list)
    passed_checks: list[str] = Field(default_factory=list)


class FormattedCode(BaseModel):
    """Output model for Code Formatter Agent."""

    formatted_function: str
    formatted_tests: str
    imports_block: str
    file_header: str = Field(description="Module docstring and metadata")
    changes_made: list[str] = Field(description="Formatting changes applied")



"""
Account Sharing Route Functions

This module provides account sharing and access management functions.

Functions:
    get_account_accesslist: Get account access list
    get_oauth_account_accesslist: Get OAuth account access list
    share_account: Share account with users/groups (v2 API)
    share_oauth_account: Share OAuth account
    share_account_v1: Legacy account sharing (v1 API)
    generate_share_account_v1_payload: Generate v1 API sharing payload
    generate_share_account_v2_payload: Generate v2 API sharing payload
    generate_share_account_payload: Orchestrator function for payload generation

Classes:
    ShareAccount: Abstract base class for sharing enums
    ShareAccount_AccessLevel: v2 API sharing permissions
    ShareAccount_V1_AccessLevel: v1 API sharing permissions (legacy)
"""

from enum import Enum
from typing import Optional, Union

import httpx
from dc_logger.decorators import LogDecoratorConfig, log_call

from ...auth import DomoAuth
from ...base.base import DomoEnumMixin
from ...client import (
    get_data as gd,
    response as rgd,
)
from ...utils.logging import ResponseGetDataProcessor
from .exceptions import AccountSharing_Error


class ShareAccount(DomoEnumMixin):
    """Abstract base class for sharing enums."""


class ShareAccount_V1_AccessLevel(ShareAccount, Enum):
    """Legacy v1 API sharing permissions (users only)."""

    OWNER = "OWNER"
    CAN_EDIT = "WRITE"
    CAN_VIEW = "READ"

    default = "READ"


class ShareAccount_AccessLevel(ShareAccount, Enum):
    """v2 API sharing permissions (users and groups)."""

    OWNER = "OWNER"
    CAN_SHARE = "CAN_SHARE"
    CAN_VIEW = "CAN_VIEW"
    CAN_EDIT = "CAN_EDIT"
    NO_ACCESS = "NONE"

    default = "CAN_VIEW"


# Payload Generation Functions


def generate_share_account_v1_payload(
    user_id: int,
    access_level: Union[ShareAccount_V1_AccessLevel, str],
) -> dict:
    """Generate v1 API sharing payload for users only.

    V1 API limitations:
    - Only supports sharing with users (no groups)
    - Limited permission set: READ, WRITE, OWNER

    Args:
        user_id: ID of the user to share with
        access_level: Access level (ShareAccount_V1_AccessLevel enum or string)

    Returns:
        Dictionary payload for v1 share API

    Example:
        >>> payload = generate_share_account_v1_payload(
        ...     user_id=12345,
        ...     access_level=ShareAccount_V1_AccessLevel.CAN_VIEW
        ... )
        >>> # Returns: {"type": "USER", "id": 12345, "permissions": ["READ"]}
    """
    if isinstance(access_level, str):
        access_level = ShareAccount_V1_AccessLevel.get(access_level)

    return {"type": "USER", "id": int(user_id), "permissions": [access_level.value]}


def generate_share_account_v2_payload(
    access_level: Union[ShareAccount_AccessLevel, str],
    user_id: int | None = None,
    group_id: int | None = None,
) -> dict:
    """Generate v2 API sharing payload for users or groups.

    V2 API features:
    - Supports both users and groups
    - Extended permissions: CAN_VIEW, CAN_EDIT, CAN_SHARE, OWNER, NO_ACCESS

    Args:
        access_level: Access level (ShareAccount_AccessLevel enum or string)
        user_id: ID of the user to share with (mutually exclusive with group_id)
        group_id: ID of the group to share with (mutually exclusive with user_id)

    Returns:
        Dictionary payload for v2 share API

    Raises:
        ValueError: If neither or both user_id and group_id are provided

    Example:
        >>> payload = generate_share_account_v2_payload(
        ...     access_level=ShareAccount_AccessLevel.CAN_VIEW,
        ...     user_id=12345
        ... )
        >>> # Returns: {"type": "USER", "id": "12345", "accessLevel": "CAN_VIEW"}
    """
    if not user_id and not group_id:
        raise ValueError("Must provide either user_id or group_id")

    if user_id and group_id:
        raise ValueError("Cannot provide both user_id and group_id")

    if isinstance(access_level, str):
        access_level = ShareAccount_AccessLevel.get(access_level)

    if user_id:
        return {"type": "USER", "id": str(user_id), "accessLevel": access_level.value}

    return {"type": "GROUP", "id": int(group_id), "accessLevel": access_level.value}


def generate_share_account_payload(
    access_level: Union[ShareAccount_AccessLevel, ShareAccount_V1_AccessLevel, str],
    user_id: int | None = None,
    group_id: int | None = None,
    use_v1_api: bool = False,
) -> dict:
    """Orchestrator function to generate appropriate sharing payload.

    This function determines which payload generation function to use based on:
    - The access_level type (v1 or v2 enum)
    - The use_v1_api flag
    - Whether a group_id is provided (forces v2)

    Args:
        access_level: Access level enum or string
        user_id: ID of the user to share with
        group_id: ID of the group to share with (forces v2 API)
        use_v1_api: Force use of v1 API (ignored if group_id provided)

    Returns:
        Dictionary payload for appropriate share API

    Raises:
        ValueError: If group_id provided with v1 access level or use_v1_api=True

    Example:
        >>> # V2 with user
        >>> payload = generate_share_account_payload(
        ...     access_level=ShareAccount_AccessLevel.CAN_VIEW,
        ...     user_id=12345
        ... )

        >>> # V2 with group (auto-detects v2 needed)
        >>> payload = generate_share_account_payload(
        ...     access_level=ShareAccount_AccessLevel.CAN_VIEW,
        ...     group_id=67890
        ... )

        >>> # V1 explicitly
        >>> payload = generate_share_account_payload(
        ...     access_level=ShareAccount_V1_AccessLevel.CAN_VIEW,
        ...     user_id=12345
        ... )
    """
    # Determine if we need v1 or v2
    is_v1_enum = isinstance(access_level, ShareAccount_V1_AccessLevel)

    # Force v2 if group_id provided
    if group_id:
        if is_v1_enum:
            raise ValueError("Cannot share with groups using v1 access levels")
        if use_v1_api:
            raise ValueError("Cannot share with groups using v1 API")

        return generate_share_account_v2_payload(
            access_level=access_level, group_id=group_id
        )

    # Use v1 if explicitly requested or v1 enum provided
    if is_v1_enum or use_v1_api:
        if not user_id:
            raise ValueError("user_id required for v1 API")

        # Convert v2 enum to v1 if needed
        if not is_v1_enum:
            # Map v2 access levels to v1
            v2_to_v1_mapping = {
                "CAN_VIEW": ShareAccount_V1_AccessLevel.CAN_VIEW,
                "CAN_EDIT": ShareAccount_V1_AccessLevel.CAN_EDIT,
                "CAN_SHARE": ShareAccount_V1_AccessLevel.OWNER,
                "OWNER": ShareAccount_V1_AccessLevel.OWNER,
            }

            if isinstance(access_level, str):
                access_level = ShareAccount_AccessLevel.get(access_level)

            access_level = v2_to_v1_mapping.get(
                access_level.value, ShareAccount_V1_AccessLevel.CAN_VIEW
            )

        return generate_share_account_v1_payload(
            user_id=user_id, access_level=access_level
        )

    # Default to v2
    return generate_share_account_v2_payload(access_level=access_level, user_id=user_id)


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def get_account_accesslist(
    auth: DomoAuth,
    account_id: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Get access list for an account.

    Args:
        auth: Authentication object for API requests
        account_id: ID of the account to get access list for
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing account access list

    Raises:
        AccountSharing_Error: If access list retrieval fails
    """
    url = (
        f"https://{auth.domo_instance}.domo.com/api/data/v2/accounts/share/{account_id}"
    )

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AccountSharing_Error(
            operation="get access list", account_id=account_id, res=res
        )

    res.response = res.response["list"]

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def get_oauth_account_accesslist(
    auth: DomoAuth,
    account_id: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Get access list for an OAuth account.

    Args:
        auth: Authentication object for API requests
        account_id: ID of the OAuth account to get access list for
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing OAuth account access list

    Raises:
        AccountSharing_Error: If OAuth access list retrieval fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/data/v2/accounts/templates/{account_id}/share"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AccountSharing_Error(
            operation="get OAuth access list", account_id=account_id, res=res
        )

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def share_account(
    auth: DomoAuth,
    account_id: str,
    share_payload: dict | None = None,
    access_level: Optional[
        Union[ShareAccount_AccessLevel, ShareAccount_V1_AccessLevel, str]
    ] = None,
    user_id: int | None = None,
    group_id: int | None = None,
    use_v1_api: bool = False,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Share account with users/groups using v2 API.

    Note: This uses the v2 API which should be deployed to all Domo instances as of DP24.

    Two ways to call this function:
    1. Pass pre-built share_payload (legacy mode)
    2. Pass access_level with user_id or group_id (generates payload automatically)

    Args:
        auth: Authentication object for API requests
        account_id: ID of the account to share
        share_payload: Pre-built sharing payload (optional, overrides other params)
        access_level: Access level enum or string (required if share_payload not provided)
        user_id: ID of user to share with (mutually exclusive with group_id)
        group_id: ID of group to share with (mutually exclusive with user_id)
        use_v1_api: Force v1 API usage (ignored if group_id provided)
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object confirming sharing operation

    Raises:
        ValueError: If neither share_payload nor access_level provided
        Account_AlreadyShared_Error: If account is already shared with target
        Account_Share_Error: If sharing operation fails

    Examples:
        >>> # Using payload generator (new way)
        >>> await share_account(
        ...     auth=auth,
        ...     account_id="123",
        ...     access_level=ShareAccount_AccessLevel.CAN_VIEW,
        ...     user_id=456
        ... )

        >>> # Using pre-built payload (legacy way)
        >>> payload = generate_share_account_v2_payload(
        ...     access_level=ShareAccount_AccessLevel.CAN_VIEW,
        ...     group_id=789
        ... )
        >>> await share_account(
        ...     auth=auth,
        ...     account_id="123",
        ...     share_payload=payload
        ... )
    """
    # Generate payload if not provided
    if not share_payload:
        if not access_level:
            raise ValueError("Must provide either share_payload or access_level")

        share_payload = generate_share_account_payload(
            access_level=access_level,
            user_id=user_id,
            group_id=group_id,
            use_v1_api=use_v1_api,
        )

    # Route to appropriate API endpoint
    if use_v1_api or (
        access_level and isinstance(access_level, ShareAccount_V1_AccessLevel)
    ):
        # Use v1 endpoint
        url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}/share"
    else:
        # Use v2 endpoint
        url = f"https://{auth.domo_instance}.domo.com/api/data/v2/accounts/share/{account_id}"

    method = "PUT"

    print(share_payload, type(share_payload))
    res = await gd.get_data(
        auth=auth,
        url=url,
        method=method,
        body=share_payload,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    if res.status == 500 and res.response == "Internal Server Error":
        raise AccountSharing_Error(
            account_id=account_id,
            message=f"{res.response} - User may already have access to account",
            res=res,
            operation=method,
        )

    if not res.is_success:
        raise AccountSharing_Error(account_id=account_id, res=res, operation=method)

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def share_oauth_account(
    auth: DomoAuth,
    account_id: str,
    share_payload: dict | None = None,
    access_level: Union[ShareAccount_AccessLevel, str] | None = None,
    user_id: int | None = None,
    group_id: int | None = None,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Share OAuth account with users/groups.

    Two ways to call this function:
    1. Pass pre-built share_payload (legacy mode)
    2. Pass access_level with user_id or group_id (generates payload automatically)

    Args:
        auth: Authentication object for API requests
        account_id: ID of the OAuth account to share
        share_payload: Pre-built sharing payload (optional, overrides other params)
        access_level: Access level enum or string (required if share_payload not provided)
        user_id: ID of user to share with (mutually exclusive with group_id)
        group_id: ID of group to share with (mutually exclusive with user_id)
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object confirming sharing operation

    Raises:
        ValueError: If neither share_payload nor access_level provided
        Account_Share_Error: If OAuth sharing operation fails

    Examples:
        >>> # Using payload generator
        >>> await share_oauth_account(
        ...     auth=auth,
        ...     account_id="123",
        ...     access_level=ShareAccount_AccessLevel.CAN_VIEW,
        ...     group_id=789
        ... )
    """
    # Generate payload if not provided
    if not share_payload:
        if not access_level:
            raise ValueError("Must provide either share_payload or access_level")

        share_payload = generate_share_account_v2_payload(
            access_level=access_level,
            user_id=user_id,
            group_id=group_id,
        )

    url = f"https://{auth.domo_instance}.domo.com/api/data/v2/accounts/templates/share/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=share_payload,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AccountSharing_Error(
            account_id=account_id,
            res=res,
        )

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def share_account_v1(
    auth: DomoAuth,
    account_id: str,
    share_payload: dict | None = None,
    access_level: Union[ShareAccount_V1_AccessLevel, str] | None = None,
    user_id: int | None = None,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Share account using legacy v1 API (users only).

    Note: V1 API allows sharing with users ONLY. It does not support sharing with groups
    and has a more limited set of share rights (owner or read). See ShareAccount_V1_AccessLevel
    vs ShareAccount_AccessLevel for differences.

    Two ways to call this function:
    1. Pass pre-built share_payload (legacy mode)
    2. Pass access_level with user_id (generates payload automatically)

    Args:
        auth: Authentication object for API requests
        account_id: ID of the account to share
        share_payload: Pre-built sharing payload (optional, overrides other params)
        access_level: Access level enum or string (required if share_payload not provided)
        user_id: ID of user to share with (required if share_payload not provided)
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object confirming sharing operation

    Raises:
        ValueError: If neither share_payload nor (access_level and user_id) provided
        Account_AlreadyShared_Error: If account is already shared with user
        Account_Share_Error: If sharing operation fails

    Examples:
        >>> # Using payload generator
        >>> await share_account_v1(
        ...     auth=auth,
        ...     account_id="123",
        ...     access_level=ShareAccount_V1_AccessLevel.CAN_VIEW,
        ...     user_id=456
        ... )
    """
    # Generate payload if not provided
    if not share_payload:
        if not access_level or not user_id:
            raise ValueError(
                "Must provide either share_payload or (access_level and user_id)"
            )

        share_payload = generate_share_account_v1_payload(
            user_id=user_id,
            access_level=access_level,
        )

    url = (
        f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}/share"
    )

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=share_payload,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    if res.status == 500 and res.response == "Internal Server Error":
        raise AccountSharing_Error(
            account_id=account_id,
            message=f"{res.response} - User may already have access to account",
            res=res,
        )

    if not res.is_success:
        raise AccountSharing_Error(
            account_id=account_id,
            res=res,
        )

    return res



"""
OAuth Account Route Functions

This module provides OAuth-specific account functions.

Functions:
    get_oauth_accounts: Retrieve all OAuth accounts user has access to
    get_oauth_account_by_id: Retrieve specific OAuth account by ID
"""

from typing import Union

import httpx
from dc_logger.decorators import LogDecoratorConfig, log_call

from ...auth import DomoAuth
from ...client import (
    get_data as gd,
    response as rgd,
)
from ...utils.logging import ResponseGetDataProcessor
from .exceptions import Account_GET_Error, AccountNoMatchError


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def get_oauth_accounts(
    auth: DomoAuth,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Retrieve all OAuth accounts the user has access to.

    Args:
        auth: Authentication object for API requests
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing OAuth account list

    Raises:
        Account_GET_Error: If OAuth account retrieval fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/templates/user/extended"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise Account_GET_Error(res=res)
    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def get_oauth_account_by_id(
    auth: DomoAuth,
    account_id: Union[int, str],
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Retrieve a specific OAuth account by ID.

    Note: This function retrieves all OAuth accounts and filters to the selected one,
    as there doesn't appear to be a direct API for retrieving a single OAuth account.

    Args:
        auth: Authentication object for API requests
        account_id: The ID of the OAuth account to retrieve
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing OAuth account metadata

    Raises:
        AccountNoMatchError: If OAuth account is not found
        Account_GET_Error: If OAuth account retrieval fails
    """
    res = await get_oauth_accounts(
        auth=auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        parent_class=parent_class,
        session=session,
        return_raw=return_raw,
    )

    if return_raw:
        return res

    # Convert account_id to int for comparison if it's a string
    target_id = int(account_id) if isinstance(account_id, str) else account_id
    res.response = next((obj for obj in res.response if obj["id"] == target_id), None)

    if not res.response:
        raise AccountNoMatchError(account_id=str(account_id), res=res)

    return res



"""DomoUser module for managing Domo users."""  # noqa: N999

__all__ = [
    "domo_default_img",
    "DomoUser",
    "DomoUsers",
    "DomoUser_NoSearchError",
    # User Route Exceptions
    "User_GET_Error",
    "User_CRUD_Error",
    "SearchUserNotFoundError",
    "UserSharing_Error",
    "DeleteUserError",
    "UserAttributes_GET_Error",
    "UserAttributes_CRUD_Error",
    "ResetPasswordPasswordUsedError",
    "DownloadAvatar_Error",
]

import asyncio
import datetime as dt
from dataclasses import dataclass, field
from typing import Any, Union

import httpx
from dc_logger.decorators import log_call

from ..auth import DomoAuth
from ..base.entities import DomoEntity, DomoManager
from ..base.exceptions import ClassError, DomoError
from ..client.response import ResponseGetData
from ..routes import user as user_routes
from ..routes.instance_config import sso as sso_routes
from ..routes.user import UserProperty
from ..routes.user.exceptions import (
    DeleteUserError,
    DownloadAvatar_Error,
    ResetPasswordPasswordUsedError,
    SearchUserNotFoundError,
    User_CRUD_Error,
    User_GET_Error,
    UserAttributes_CRUD_Error,
    UserAttributes_GET_Error,
    UserSharing_Error,
)
from ..utils.convert import convert_epoch_millisecond_to_datetime, test_valid_email
from ..utils.images import Image, ImageUtils, are_same_image
from ..utils.logging import get_colored_logger

# User route exceptions are now imported from ..routes.user.exceptions

logger = get_colored_logger()


class CreateUser_MissingRoleError(ClassError):  # noqa: N801
    """Exception raised when role_id is missing during user creation."""

    def __init__(self, domo_instance, email_address):
        super().__init__(
            domo_instance=domo_instance,
            message=f"Error creating user {email_address}: missing role_id parameter",
        )


class DomoUser_NoSearchError(ClassError):  # noqa: N801
    """Exception raised when user search operations fail."""

    def __init__(
        self,
        cls_instance,
        message: str = "No users found matching search criteria",
        domo_instance: str = None,
    ):
        super().__init__(
            cls_instance=cls_instance,
            domo_instance=domo_instance,
            message=message,
        )


default_img_bytes = b""  # Placeholder for actual byte data

# Update type annotations to use Optional

domo_default_img = None  # Placeholder for the default image


@dataclass(eq=False)
class DomoUser(DomoEntity):
    """A class for interacting with a Domo User"""

    id: str
    display_name: str | None = None
    email_address: str | None = None
    role_id: int | None = None  # Updated to match expected type
    department: str | None = None
    title: str | None = None
    avatar_key: str | None = None
    avatar: Image.Image | None = None
    password: str | None = field(repr=False, default=None)

    phone_number: str | None = None
    web_landing_page: str | None = None
    web_mobile_landing_page: str | None = None
    employee_id: str | None = None
    employee_number: str | None = None
    hire_date: str | None = None
    reports_to: str | None = None

    publisher_domain: str | None = None
    subscriber_domain: str | None = None
    virtual_user_id: str | None = None

    created_dt: dt.datetime | None = None
    last_activity_dt: dt.datetime | None = None

    custom_attributes: dict = field(default_factory=dict)

    domo_api_clients: list[Any] | None = None
    domo_access_tokens: list[Any] | None = None

    Role: Any | None = None  # DomoRole
    ApiClients: Any | None = None  # DomoApiClients

    def __post_init__(self):
        from .DomoInstanceConfig.api_client import ApiClients

        self.id = str(self.id)

        self.ApiClients = ApiClients.from_parent(auth=self.auth, parent=self)

    @property
    def entity_type(self) -> str:
        return "USER"

    @property
    def display_url(self) -> str:
        """Generate the URL to display this user in the Domo admin interface."""
        return f"https://{self.auth.domo_instance}.domo.com/admin/people/{self.id}"

    @classmethod
    def from_dict(cls, auth, obj: dict):
        return cls(
            auth=auth,
            id=str(obj.get("id") or obj.get("userId")),
            display_name=obj.get("displayName"),
            title=obj.get("title"),
            department=obj.get("department"),
            email_address=obj.get("emailAddress") or obj.get("email"),
            role_id=obj.get("roleId"),
            avatar_key=obj.get("avatarKey"),
            phone_number=obj.get("phoneNumber"),
            web_landing_page=obj.get("webLandingPage"),
            web_mobile_landing_page=obj.get("webMobileLandingPage"),
            employee_id=obj.get("employeeId"),
            employee_number=obj.get("employeeNumber"),
            hire_date=obj.get("hireDate"),
            reports_to=obj.get("reportsTo"),
            created_dt=convert_epoch_millisecond_to_datetime(obj.get("created")),
            last_activity_dt=convert_epoch_millisecond_to_datetime(
                obj.get("lastActivity"),
            ),
            raw=obj,
        )

    @classmethod
    def from_virtual_dict(cls, auth, obj: dict):
        """Create DomoUser instance from virtual user JSON response.

        Virtual users are created when content is published across instances.
        """
        return cls(
            id=obj["id"],
            auth=auth,
            publisher_domain=obj.get("publisherDomain"),
            subscriber_domain=obj.get("subscriberDomain"),
            virtual_user_id=obj.get("virtualUserId"),
            raw=obj,
        )

    @classmethod
    def from_bootstrap_dict(cls, auth, obj):
        return cls(
            id=obj["id"],
            display_name=obj.get("displayName"),
            auth=auth,
            raw=obj,
        )

    async def get_role(
        self: "DomoUser",
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient | None = None,
    ):
        from .DomoInstanceConfig.role import DomoRole

        self.Role = await DomoRole.get_by_id(
            role_id=str(self.role_id),
            auth=self.auth,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
        )

        return self.Role

    @classmethod
    async def get_by_id(
        cls,
        auth: DomoAuth,
        user_id: str,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient | None = None,
    ):
        """Retrieve a Domo user by ID.

        Args:
            auth: Authentication object
            entity_id: User ID to retrieve
            return_raw: If True, return raw API response instead of DomoUser instance
            debug_api: Enable debug output for API call
            debug_num_stacks_to_drop: Number of stack frames to drop in logging
            session: Optional httpx session for connection pooling

        Returns:
            DomoUser instance or ResponseGetData if return_raw=True

        Raises:
            User_GET_Error: If user retrieval fails
        """

        res = await user_routes.get_by_id(
            auth=auth,
            user_id=user_id,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
            parent_class=cls.__name__,
        )

        if return_raw:
            return res

        if not res.is_success:
            return None

        domo_user = cls.from_dict(obj=res.response, auth=auth)

        try:
            await domo_user.get_role(
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                session=session,
            )
        except DomoError as e:
            print(e)

        return domo_user

    @classmethod
    async def get_entity_by_id(cls, entity_id: str, auth: DomoAuth, **kwargs):
        return await cls.get_by_id(user_id=entity_id, auth=auth, **kwargs)

    async def download_avatar(
        self,
        pixels: int = 300,
        folder_path="./images",
        img_name=None,  # will default to user_id
        auth: DomoAuth | None = None,
        is_download_image: bool = True,  # option to prevent downloading the image file
        debug_api: bool = False,
        return_raw: bool = False,
    ):
        """downloads a user's avatar to a folder
        and returns the byte representation of the image
        """
        auth = auth or self.auth

        res = await user_routes.download_avatar(
            auth=self.auth,
            user_id=self.id,
            pixels=pixels,
            folder_path=folder_path,
            img_name=img_name,
            is_download_image=is_download_image,
            debug_api=debug_api,
        )

        if return_raw:
            return res

        self.avatar = ImageUtils.from_bytestr(data=res.response)

        return self.avatar

        # return res.response

    async def update_properties(
        self,
        property_ls: list[
            UserProperty
        ],  # use the UserProperty class to define a list of user properties to update, see user route documentation to see a list of UserProperty_Types that can be updated
        return_raw: bool = False,
        auth: DomoAuth | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
    ):
        auth = auth or self.auth

        res = await user_routes.update_user(
            auth=auth,
            user_id=self.id,
            user_property_ls=property_ls,
            debug_api=debug_api,
            session=session,
        )
        if return_raw:
            return res

        # Update self using from_dict pattern
        updated_user = self.from_dict(auth=auth, obj=res.response)
        # Copy updated attributes back to self
        for key, value in updated_user.__dict__.items():
            if key not in ["auth"]:  # Don't overwrite auth
                setattr(self, key, value)

        # Update self using from_dict pattern
        if res.response:
            updated_user = self.from_dict(auth=auth, obj=res.response)
            # Copy updated attributes back to self
            for key, value in updated_user.__dict__.items():
                if key not in ["auth"]:  # Don't overwrite auth
                    setattr(self, key, value)

            return updated_user

    async def set_user_landing_page(
        self,
        page_id: str,
        id: str,
        auth: DomoAuth | None = None,
        debug_api: bool = False,
    ):
        res = await user_routes.set_user_landing_page(
            auth=auth or self.auth,
            page_id=page_id,
            user_id=self.id or id,
            debug_api=debug_api,
        )

        return res

    @classmethod
    async def create(
        cls,
        auth: DomoAuth,
        display_name,
        email_address,
        role: Any,  # DomoRole - will use default role if none provided
        password: str | None = None,
        send_password_reset_email: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient | None = None,
    ):
        """class method that creates a new Domo user"""

        if not role:
            from .DomoInstanceConfig.role import DomoRoles

            role = await DomoRoles(auth=auth).get_default_role(session=session)

        res = await user_routes.create_user(
            auth=auth,
            display_name=display_name,
            email_address=email_address,
            role_id=role.id,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        domo_user = await DomoUser.get_by_id(
            auth=auth,
            user_id=res.response.get("id") or res.response.get("userId"),
            session=session,
        )

        if password and domo_user:
            await domo_user.reset_password(
                new_password=password, session=session, debug_api=debug_api
            )

        elif send_password_reset_email and domo_user:
            await domo_user.request_password_reset(
                domo_instance=auth.domo_instance,
                email=domo_user.email_address,
                session=session,
                debug_api=debug_api,
            )

        return domo_user

    async def delete(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        debug_num_stacks_to_drop=2,
        parent_class=None,
    ):
        res = await user_routes.delete_user(
            auth=self.auth,
            user_id=self.id,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=parent_class,
        )

        return res

    async def reset_password(
        self: "DomoUser",
        new_password: str,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        debug_num_stacks_to_drop: int = 2,
    ):
        """reset your password, will respect password restrictions set up in the Domo UI"""

        res = await user_routes.reset_password(
            auth=self.auth,
            user_id=self.id,
            new_password=new_password,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        self.password = new_password

        return res

    @classmethod
    async def request_password_reset(
        cls,
        domo_instance: str,
        email: str,
        locale: str = "en-us",
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        debug_num_stacks_to_drop: int = 2,
    ):
        """request password reset email.  Note: does not require authentication."""

        return await user_routes.request_password_reset(
            domo_instance=domo_instance,
            email=email,
            locale=locale,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=cls.__name__,
        )

    async def upload_avatar(
        self,
        avatar: Image.Image,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        return_raw: bool = False,
    ):
        avatar = ImageUtils.crop_square(avatar)

        res = await user_routes.upload_avatar(
            auth=self.auth,
            user_id=self.id,
            img_bytestr=ImageUtils.to_bytes(avatar),
            img_type=avatar.format,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            session=session,
        )

        if return_raw:
            return res

        await asyncio.sleep(2)

        return await self.download_avatar(debug_api=debug_api)

    async def upsert_avatar(
        self,
        avatar: Image.Image,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        return_raw: bool = False,
    ):
        avatar = ImageUtils.crop_square(avatar)

        res = "images are the same"
        if not are_same_image(domo_default_img, avatar):
            res = await user_routes.upload_avatar(
                auth=self.auth,
                user_id=self.id,
                img_bytestr=ImageUtils.to_bytes(avatar),
                img_type=avatar.format,
                debug_api=debug_api,
                parent_class=self.__class__.__name__,
                session=session,
            )

            if return_raw:
                return res

        await asyncio.sleep(2)

        return await self.download_avatar(debug_api=debug_api)

    async def toggle_direct_signon_access(
        self: "DomoUser",
        is_enable_direct_signon: bool = None,
        note: str | None = None,  # add a note to the sso entry
        session: httpx.AsyncClient | None = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
    ):
        response = ""
        if is_enable_direct_signon is not None:
            res = await sso_routes.toggle_user_direct_signon_access(
                auth=self.auth,
                user_id_ls=[self.id],
                is_enable_direct_signon=is_enable_direct_signon,
                session=session,
                debug_api=debug_api,
                parent_class=self.__class__.__name__,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                return_raw=return_raw,
            )

            response = res.response

        if note is not None:
            note_res = await sso_routes.add_note_to_user_direct_signon(
                auth=self.auth,
                user_id_ls=[self.id],
                note=note,
                session=session,
                debug_api=debug_api,
                parent_class=self.__class__.__name__,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                return_raw=return_raw,
            )

            response += note_res.response

        if return_raw:
            return res

        res.response = response
        return res

    async def get_api_clients(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
    ):
        """
        retrieves Client_IDs for this user (assuming the authenticated user has manage rights).
        Note : the values will be masked, raw text values can only be retrieved via the UI
        """

        from .DomoInstanceConfig.api_client import ApiClients

        api_clients = ApiClients(auth=self.auth)

        res = await api_clients.get(
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        if return_raw:
            return res

        domo_clients = (
            [
                domo_client
                for domo_client in res.response
                if domo_client.domo_user.id == self.id  # type: ignore
            ]
            if isinstance(res, ResponseGetData)
            else []
        )

        if not domo_clients:
            print(
                f"Domo User {self.id} - {self.display_name} does not have any Client_IDs"
            )
            return False

        self.domo_api_clients = domo_clients

        return self.domo_api_clients

    async def get_access_tokens(
        self,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient | None = None,
        return_raw: bool = False,
    ):
        from .DomoInstanceConfig import access_token as dmat

        domo_config = dmat.DomoAccessTokens(auth=self.auth)
        domo_tokens = await domo_config.get(
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        if return_raw:
            return domo_tokens

        domo_tokens = [
            domo_token for domo_token in domo_tokens if domo_token.owner.id == self.id
        ]

        if not domo_tokens:
            print(
                f"Domo User {self.id} - {self.display_name} does not have any access tokens"
            )
            return []

        self.domo_access_tokens = domo_tokens

        return self.domo_access_tokens


@dataclass
class DomoUsers(DomoManager):
    """a class for searching for Users"""

    auth: DomoAuth = field(repr=False)
    users: list[DomoUser] = field(default_factory=list)
    virtual_users: list[DomoUser] = field(default_factory=list)

    @classmethod
    def _users_to_domo_user(cls, user_ls, auth: DomoAuth):
        return [DomoUser.from_dict(auth=auth, obj=obj) for obj in user_ls]

    @classmethod
    def _users_to_virtual_user(cls, user_ls, auth: DomoAuth):
        return [DomoUser.from_virtual_dict(auth=auth, obj=obj) for obj in user_ls]

    @staticmethod
    def _util_match_domo_users_to_emails(
        domo_users: list[DomoUser], user_email_ls: list[str]
    ) -> list[DomoUser]:
        """pass in an array of user emails to match against an array of Domo User"""

        return [
            domo_user
            for domo_user in domo_users
            if domo_user.email_address
            and domo_user.email_address.lower()
            in [email.lower() for email in user_email_ls]
        ]

    @staticmethod
    def _util_match_users_obj_to_emails(
        user_ls: list[dict], user_email_ls: list[str]
    ) -> list:
        """pass in an array of user emails to match against an array of Domo User"""

        return [
            obj
            for obj in user_ls
            if obj.get("emailAddress", "").lower()
            in [email.lower() for email in user_email_ls]
        ]

    async def get(
        self,
        *args,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient | None = None,
        **kwargs,
    ) -> list[DomoUser]:
        """retrieves all users from Domo"""

        res = await user_routes.get_all_users(
            auth=self.auth,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

        if return_raw:
            return res

        self.users = self._users_to_domo_user(user_ls=res.response, auth=self.auth)
        return self.users

    async def search_by_email(
        self,
        email: Union[str, list[str]],
        only_allow_one: bool = True,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
        suppress_no_results_error: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> Union[list[DomoUser], DomoUser, ResponseGetData, bool]:
        emails = [email] if isinstance(email, str) else email

        try:
            res = await user_routes.search_users_by_email(
                user_email_ls=emails,
                auth=self.auth,
                return_raw=return_raw,
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                parent_class=self.__class__.__name__,
                session=session,
            )

        except SearchUserNotFoundError as e:
            if suppress_no_results_error:
                return []

            raise e from e

        if return_raw:
            return res

        domo_users = self._users_to_domo_user(res.response, auth=self.auth)

        if not only_allow_one:
            return domo_users

        domo_users = self._util_match_domo_users_to_emails(domo_users, emails)

        if not domo_users:
            raise DomoUser_NoSearchError(
                cls_instance=self,
                message=f"unable to find {','.join(emails)}",
                domo_instance=self.auth.domo_instance,
            )

        return domo_users[0]

    async def search_by_id(
        self,
        user_ids: list[str],  # can search for one or multiple users
        suppress_no_results_error: bool = False,
        only_allow_one: bool = True,
        debug_num_stacks_to_drop=2,
        debug_api: bool = False,
        return_raw: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> Union[list[DomoUser], DomoUser, ResponseGetData, bool]:
        res = None

        try:
            res = await user_routes.search_users_by_id(
                return_raw=return_raw,
                user_ids=user_ids,
                debug_api=debug_api,
                auth=self.auth,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                parent_class=self.__class__.__name__,
                session=session,
            )

        except SearchUserNotFoundError as e:
            if suppress_no_results_error:
                print(e)

                if only_allow_one:
                    return False
                return []

            raise e from e

        if return_raw:
            return res

        domo_users = self._users_to_domo_user(user_ls=res.response, auth=self.auth)

        if only_allow_one:
            return domo_users[0]

        return domo_users

    async def get_virtual_user_by_subscriber_instance(
        self,
        subscriber_instance_ls: str,
        debug_api: bool = False,
        return_raw: bool = False,
    ):
        res = await user_routes.search_virtual_user_by_subscriber_instance(
            auth=self.auth,
            subscriber_instance_ls=subscriber_instance_ls,
            debug_api=debug_api,
        )

        if return_raw:
            return res

        domo_users = self._users_to_virtual_user(res.response, auth=self.auth)
        self.virtual_users = domo_users
        return domo_users

    @log_call(action_name="class")
    async def upsert(
        self,
        email_address: str,
        display_name: str | None = None,
        role: Any | None = None,  # DomoRole
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient | None = None,
    ) -> DomoUser:
        """Upsert a Domo user - update if exists, create if not.

        Args:
            email_address: User email address (required)
            display_name: User display name (optional)
            role: DomoRole object (optional, will use default if not provided on creation)
            debug_api: Enable debug output for API calls
            debug_num_stacks_to_drop: Number of stack frames to drop in logging
            session: Optional httpx session for connection pooling

        Returns:
            DomoUser: The created or updated user

        Raises:
            ValidationError: If email address is invalid
        """
        test_valid_email(email_address)

        try:
            # Type narrowing: when only_allow_one=True and successful, returns DomoUser
            domo_user: DomoUser = await self.search_by_email(
                email=email_address,
                only_allow_one=True,
                debug_api=debug_api,
                session=session,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
            )

            await logger.info(f"domo_user found {domo_user.id}")

            # Type guard to ensure we have a DomoUser instance
            if not isinstance(domo_user, DomoUser):
                raise ValueError(f"Expected DomoUser, got {type(domo_user)}")

            property_ls = []
            if display_name:
                property_ls.append(
                    user_routes.UserProperty(
                        user_routes.UserProperty_Type.display_name, display_name
                    )
                )

            if role and hasattr(role, "id") and role.id:
                property_ls.append(
                    user_routes.UserProperty(
                        user_routes.UserProperty_Type.role_id, role.id
                    )
                )

            if property_ls:
                await logger.info("Updating user properties for existing user.")
                await domo_user.update_properties(
                    property_ls=property_ls,
                    debug_api=debug_api,
                    session=session,
                )
            return domo_user

        except (SearchUserNotFoundError, DomoUser_NoSearchError):
            # User doesn't exist, create new one

            await logger.info("User not found, creating new user.")

            created_user = await DomoUser.create(
                display_name=display_name
                or f"{email_address} - via dl {dt.date.today()}",
                email_address=email_address,
                debug_api=debug_api,
                session=session,
                auth=self.auth,
                role=role,
            )

            # Type guard for created user
            if created_user is None:
                raise ValueError(f"Failed to create user for email: {email_address}")

            await self.get()

            return created_user



"""Dataset core CRUD operations."""

import httpx
from dc_logger.decorators import LogDecoratorConfig, log_call

from ...auth import DomoAuth
from ...client import (
    get_data as gd,
    response as rgd,
)
from ...utils.logging import DomoEntityExtractor, DomoEntityResultProcessor
from .exceptions import Dataset_CRUD_Error, Dataset_GET_Error, DatasetNotFoundError


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def get_dataset_by_id(
    dataset_id: str,  # dataset id from URL
    auth: DomoAuth | None = None,  # requires full authentication
    debug_api: bool = False,  # for troubleshooting API request
    session: httpx.AsyncClient | None = None,
    parent_class: str | None = None,
    debug_num_stacks_to_drop=1,
    context=None,  # RouteContext from route_function decorator
) -> rgd.ResponseGetData:  # returns metadata about a dataset
    """retrieve dataset metadata"""

    url = f"https://{auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}"  # type: ignore

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if res.status == 404 and res.response == "Not Found":
        raise DatasetNotFoundError(dataset_id=dataset_id, res=res)

    if not res.is_success:
        raise Dataset_GET_Error(dataset_id=dataset_id, res=res)

    return res


def generate_create_dataset_body(
    dataset_name: str, dataset_type: str = "API", schema: dict | None = None
):
    schema = schema or {
        "columns": [
            {"type": "STRING", "name": "Friend"},
            {"type": "STRING", "name": "Attending"},
        ]
    }

    return {
        "userDefinedType": dataset_type,
        "dataSourceName": dataset_name,
        "schema": schema,
    }


@gd.route_function
async def create(
    auth: DomoAuth,
    dataset_name: str,
    dataset_type: str = "api",
    schema: dict | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str | None = None,
    session: httpx.AsyncClient | None = None,
    context=None,
):
    body = generate_create_dataset_body(
        dataset_name=dataset_name, dataset_type=dataset_type, schema=schema
    )

    url = f"https://{auth.domo_instance}.domo.com/api/data/v2/datasources"

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise Dataset_CRUD_Error(res=res)

    return res


def generate_enterprise_toolkit_body(
    dataset_name, dataset_description, datasource_type, columns_schema: list[dict]
):
    return {
        "dataSourceName": dataset_name,
        "dataSourceDescription": dataset_description,
        "dataSourceType": datasource_type,
        "schema": {"columns": columns_schema},
    }


def generate_remote_domostats_body(
    dataset_name, dataset_description, columns_schema: list[dict] = []
):
    return generate_enterprise_toolkit_body(
        dataset_name=dataset_name,
        dataset_description=dataset_description,
        columns_schema=columns_schema
        or [{"type": "STRING", "name": "Remote Domo Stats"}],
        datasource_type="ObservabilityMetrics",
    )


@gd.route_function
async def create_dataset_enterprise_tookit(
    auth: DomoAuth,
    payload: dict,  # call generate_enterprise_toolkit_body
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class=None,
    session: httpx.AsyncClient | None = None,
    context=None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/executor/v1/datasets"

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=payload,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise Dataset_CRUD_Error(res=res)

    return res


@gd.route_function
async def delete_partition_stage_1(
    auth: DomoAuth,
    dataset_id: str,
    dataset_partition_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class=None,
    session: httpx.AsyncClient | None = None,
    context=None,
):
    """Delete partition has 3 stages
    # Stage 1. This marks the data version associated with the partition tag as deleted.
    It does not delete the partition tag or remove the association between the partition tag and data version.
    There should be no need to upload an empty file – step #3 will remove the data from Adrenaline.
    # update on 9/9/2022 based on the conversation with Greg Swensen"""

    url = f"https://{auth.domo_instance}.domo.com/api/query/v1/datasources/{dataset_id}/tag/{dataset_partition_id}/data"

    res = await gd.get_data(
        auth=auth,
        method="DELETE",
        url=url,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise Dataset_CRUD_Error(dataset_id=dataset_id, res=res)

    return res


@gd.route_function
async def delete_partition_stage_2(
    auth: DomoAuth,
    dataset_id: str,
    dataset_partition_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class=None,
    session: httpx.AsyncClient | None = None,
    context=None,
):
    """This will remove the partition association so that it doesn't show up in the list call.
    Technically, this is not required as a partition against a deleted data version will not count against the 400 partition limit
    but as the current partitions api doesn't make that clear, cleaning these up will make it much easier for you to manage.
    """

    url = f"https://{auth.domo_instance}.domo.com/api/query/v1/datasources/{dataset_id}/partition/{dataset_partition_id}"

    res = await gd.get_data(
        auth=auth,
        method="DELETE",
        url=url,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise Dataset_CRUD_Error(dataset_id=dataset_id, res=res)

    return res


@gd.route_function
async def delete(
    auth: DomoAuth,
    dataset_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class=None,
    session: httpx.AsyncClient | None = None,
    context=None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}?deleteMethod=hard"

    res = await gd.get_data(
        auth=auth,
        method="DELETE",
        url=url,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise Dataset_CRUD_Error(dataset_id=dataset_id, res=res)

    return res



"""
AppDb Datastore Functions

This module provides functions for managing Domo AppDb datastores including
retrieval and creation operations.

Functions:
    get_datastores: Retrieve all datastores
    get_datastore_by_id: Retrieve a specific datastore by ID
    get_collections_from_datastore: Get collections from a specific datastore
    create_datastore: Create a new datastore
"""

__all__ = [
    "get_datastores",
    "get_datastore_by_id",
    "get_collections_from_datastore",
    "create_datastore",
]


import httpx

from ...auth import DomoAuth
from ...client import (
    get_data as gd,
    response as rgd,
)
from .exceptions import AppDb_CRUD_Error, AppDb_GET_Error, SearchAppDb_NotFound


@gd.route_function
async def get_datastores(
    auth: DomoAuth,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Retrieve all datastores.

    Args:
        auth: Authentication object containing credentials and instance info
        session: Optional httpx client session for connection reuse
        debug_api: Enable detailed API request/response logging
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output
        parent_class: Optional parent class name for debugging context
        return_raw: Return raw API response without processing

    Returns:
        ResponseGetData object containing datastores information

    Raises:
        AppDb_GET_Error: If datastores retrieval fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppDb_GET_Error(res=res)

    return res


@gd.route_function
async def get_datastore_by_id(
    auth: DomoAuth,
    datastore_id: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Retrieve a specific datastore by ID.

    Args:
        auth: Authentication object containing credentials and instance info
        datastore_id: Unique identifier for the datastore
        session: Optional httpx client session for connection reuse
        debug_api: Enable detailed API request/response logging
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output
        parent_class: Optional parent class name for debugging context
        return_raw: Return raw API response without processing

    Returns:
        ResponseGetData object containing datastore information

    Raises:
        AppDb_GET_Error: If datastore retrieval fails
        SearchAppDb_NotFound: If datastore with specified ID doesn't exist
    """
    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/{datastore_id}"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        if res.status == 404:
            raise SearchAppDb_NotFound(
                search_criteria=f"datastore_id: {datastore_id}",
                res=res,
            )
        raise AppDb_GET_Error(appdb_id=datastore_id, res=res)

    return res


@gd.route_function
async def get_collections_from_datastore(
    auth: DomoAuth,
    datastore_id: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Get collections from a specific datastore.

    Args:
        auth: Authentication object containing credentials and instance info
        datastore_id: Unique identifier for the datastore
        session: Optional httpx client session for connection reuse
        debug_api: Enable detailed API request/response logging
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output
        parent_class: Optional parent class name for debugging context
        return_raw: Return raw API response without processing

    Returns:
        ResponseGetData object containing collections information

    Raises:
        AppDb_GET_Error: If collections retrieval fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/{datastore_id}/collections"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppDb_GET_Error(appdb_id=datastore_id, res=res)

    return res


@gd.route_function
async def create_datastore(
    auth: DomoAuth,
    datastore_name: str | None = None,  # in UI shows up as appName
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Create a new datastore.

    Args:
        auth: Authentication object containing credentials and instance info
        datastore_name: Name for the new datastore (shows as appName in UI)
        session: Optional httpx client session for connection reuse
        debug_api: Enable detailed API request/response logging
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output
        parent_class: Optional parent class name for debugging context
        return_raw: Return raw API response without processing

    Returns:
        ResponseGetData object containing created datastore information

    Raises:
        AppDb_CRUD_Error: If datastore creation fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/"

    body = {"name": datastore_name}

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppDb_CRUD_Error(operation="create", res=res)

    return res



"""Page management and collection operations."""

__all__ = ["DomoPages"]

from dataclasses import dataclass, field
from typing import Any

import httpx

from ...auth import DomoAuth
from ...routes import page as page_routes
from ...utils import chunk_execution as dmce


@dataclass
class DomoPages:
    auth: DomoAuth = field(repr=False)
    pages: list[Any] = None  # DomoPage

    async def get(self, **kwargs):
        """calls get_admin_summary to retrieve all pages in an instance"""
        return await self.get_admin_summary(**kwargs)

    async def get_admin_summary(
        self,
        search_title: str = None,
        parent_page_id: int = None,
        return_raw: bool = False,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
    ):
        """use admin_summary to retrieve all pages in an instance -- regardless of user access
        NOTE: some Page APIs will not return results if page access isn't explicitly shared
        """
        from .core import DomoPage

        res = await page_routes.get_pages_adminsummary(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            search_title=search_title,
            page_parent_id=parent_page_id,
        )

        if return_raw:
            return res

        self.pages = await dmce.gather_with_concurrency(
            n=60,
            *[
                DomoPage._from_adminsummary(
                    page_obj, auth=self.auth, debug_api=debug_api, session=session
                )
                for page_obj in res.response
            ],
        )

        return self.pages



"""Route context for API request configuration.

This module provides the RouteContext dataclass for consolidating
debug and session parameters across route functions.
"""

from dataclasses import dataclass

import httpx
from dc_logger import LogLevel


@dataclass
class RouteContext:
    """Context object for route function execution.

    Consolidates debugging and session parameters that are commonly
    passed to route functions and get_data calls.

    Attributes:
        session: Optional httpx client session for connection reuse
        debug_api: Enable detailed API request/response logging
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output
        parent_class: Optional parent class name for debugging context
        log_level: Optional log level for the request
        dry_run: If True, return request parameters without executing the API call
    """

    session: httpx.AsyncClient | None = None
    debug_num_stacks_to_drop: int = None
    parent_class: str | None = None
    log_level: LogLevel | str | None = None
    debug_api: bool = None
    dry_run: bool = None

    @classmethod
    def build_context(
        cls,
        context: "RouteContext" | None = None,
        **kwargs,
    ) -> "RouteContext":
        """Build RouteContext from either existing context or individual parameters.

        This helper allows route functions to accept either a pre-built context
        or individual parameters, enabling clean signatures while maintaining
        backward compatibility.

        Args:
            context: Optional pre-built RouteContext
            **kwargs: Individual context parameters (session, debug_api, debug_num_stacks_to_drop,
                    parent_class, log_level, dry_run)

        Returns:
            RouteContext built from provided parameters

        Example:
            >>> # In a route function
            >>> def my_route(auth, *, context=None, **context_kwargs):
            ...     context = build_context(context, **context_kwargs)
            ...     res = await get_data(auth=auth, url=url, context=context)
        """
        context = context or RouteContext()

        # Update context attributes from kwargs if provided
        # For boolean values, allow False to be set explicitly

        for key, value in kwargs.items():
            if hasattr(context, key):
                if value is not None:
                    setattr(context, key, value)

        return context



"""
Domo entity system.

This package provides foundational classes for all Domo entities with support
for authentication, relationships, lineage tracking, and entity management.

Modules:
- base: Foundational classes and enhanced enums
- entities: Core Domo entity classes and managers
- relationships: Relationship system for entity interactions

The design provides a consistent interface across all Domo entity types
while supporting advanced features like lineage tracking and relationships.
"""

# Import base classes and enums
from .base import DomoBase, DomoEnum, DomoEnumMixin

# Import entities from existing entities.py module
from .entities import (
    DomoEntity,
    DomoEntity_w_Lineage,
    DomoManager,
    DomoSubEntity,
)

# Import federated entities
from .entities_federated import (
    DomoFederatedEntity,
    DomoPublishedEntity,
)

# Import relationship system
from .relationships import (
    DomoRelationship,
    DomoRelationshipController,
)

__all__ = [
    # Base classes and enums
    "DomoEnum",
    "DomoEnumMixin",
    "DomoBase",
    # Core entities
    "DomoEntity",
    "DomoEntity_w_Lineage",
    "DomoManager",
    "DomoSubEntity",
    # Federated entities
    "DomoFederatedEntity",
    "DomoPublishedEntity",
    # Relationships
    "DomoRelationship",
    "DomoRelationshipController",
]



"""DomoJupyter module - Jupyter workspace management."""

from .Account import DomoJupyter_Account
from .Content import DomoJupyter_Content
from .DataSource import DomoJupyter_DataSource
from .Jupyter import (
    DJW_InvalidClass,
    DJW_Search_Error,
    DomoJupyterWorkspace,
    DomoJupyterWorkspaces,
)

__all__ = [
    "DomoJupyter_Account",
    "DomoJupyter_Content",
    "DomoJupyter_DataSource",
    "DJW_InvalidClass",
    "DJW_Search_Error",
    "DomoJupyterWorkspace",
    "DomoJupyterWorkspaces",
]



"""
Page Access Functions

This module provides functions for managing page access control and permissions.

Functions:
    get_page_access_test: Test page access permissions for the authenticated user
    get_page_access_list: Retrieve page access list showing which users and groups have access
    add_page_owner: Add owners to multiple pages
"""

__all__ = [
    "get_page_access_test",
    "get_page_access_list",
    "add_page_owner",
]


import httpx
from dc_logger.decorators import LogDecoratorConfig, log_call

from ...auth import DomoAuth
from ...client import (
    get_data as gd,
    response as rgd,
)
from ...utils.logging import DomoEntityExtractor, DomoEntityResultProcessor
from .exceptions import (
    Page_CRUD_Error,
    Page_GET_Error,
    PageSharing_Error,
    SearchPageNotFoundError,
)


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def get_page_access_test(
    auth: DomoAuth,
    page_id: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Test page access permissions for the authenticated user.

    Args:
        auth: Authentication object containing credentials and instance info
        page_id: Unique identifier for the page
        session: Optional httpx client session for connection reuse
        debug_api: Enable detailed API request/response logging
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output
        parent_class: Optional parent class name for debugging context
        return_raw: Return raw API response without processing

    Returns:
        ResponseGetData object containing page access information

    Raises:
        Page_GET_Error: If access test fails
        SearchPageNotFoundError: If page with specified ID doesn't exist
    """
    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/pages/{page_id}/access"

    res = await gd.get_data(
        url,
        method="GET",
        auth=auth,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if return_raw:
        return res

    if not res.is_success:
        if res.status == 404:
            raise SearchPageNotFoundError(
                search_criteria=f"page_id: {page_id}",
                res=res,
            )
        raise Page_GET_Error(page_id=page_id, res=res)

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def get_page_access_list(
    auth: DomoAuth,
    page_id: str,
    is_expand_users: bool = True,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Retrieve page access list showing which users and groups have access.

    Args:
        auth: Authentication object containing credentials and instance info
        page_id: Unique identifier for the page
        is_expand_users: Whether to expand group memberships to include individual users
        session: Optional httpx client session for connection reuse
        debug_api: Enable detailed API request/response logging
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output
        parent_class: Optional parent class name for debugging context
        return_raw: Return raw API response without processing

    Returns:
        ResponseGetData object containing access list with users and groups

    Raises:
        PageSharing_Error: If access list retrieval fails
        SearchPageNotFoundError: If page with specified ID doesn't exist
    """

    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/share/accesslist/page/{page_id}?expandUsers={is_expand_users}"

    res = await gd.get_data(
        url,
        method="GET",
        auth=auth,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if return_raw:
        return res

    if not res.is_success:
        if res.status == 404:
            raise SearchPageNotFoundError(
                search_criteria=f"page_id: {page_id}",
                res=res,
            )
        raise PageSharing_Error(
            operation="retrieve access list",
            page_id=page_id,
            res=res,
        )

    res.response["explicitSharedUserCount"] = len(res.response.get("users"))
    for user in res.response.get("users"):
        user.update({"isExplicitShare": True})

    # add group members to users response
    if is_expand_users:
        group_users = [
            {**user, "isExplicitShare": False}
            for group in res.response.get("groups")
            for user in group.get("users")
        ]
        users = res.response.get("users") + [
            group_user
            for group_user in group_users
            if group_user.get("id")
            not in [user.get("id") for user in res.response.get("users")]
        ]
        res.response.update({"users": users})

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def add_page_owner(
    auth: DomoAuth,
    page_id_ls: list[int | str],
    group_id_ls: list[int | str] | None = None,
    user_id_ls: list[int | str] | None = None,
    note: str = "",
    send_email: bool = False,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Add owners to multiple pages.

    Args:
        auth: Authentication object containing credentials and instance info
        page_id_ls: list of page IDs to add owners to
        group_id_ls: Optional list of group IDs to add as owners
        user_id_ls: Optional list of user IDs to add as owners
        note: Optional note to include with ownership changes
        send_email: Whether to send notification email
        session: Optional httpx client session for connection reuse
        debug_api: Enable detailed API request/response logging
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output
        parent_class: Optional parent class name for debugging context
        return_raw: Return raw API response without processing

    Returns:
        ResponseGetData object containing ownership update result

    Raises:
        Page_CRUD_Error: If adding page owners fails
    """
    page_id_ls = [str(ele) for ele in page_id_ls]
    group_id_ls = group_id_ls or []
    user_id_ls = user_id_ls or []

    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/pages/bulk/owners"
    owners = []
    for group in group_id_ls:
        owners.append({"id": str(group), "type": "GROUP"})
    for user in user_id_ls:
        owners.append({"id": str(user), "type": "USER"})

    body = {
        "pageIds": page_id_ls,
        "owners": owners,
        "note": note,
        "sendEmail": send_email,
    }

    res = await gd.get_data(
        auth=auth,
        method="PUT",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise Page_CRUD_Error(
            operation="add owners",
            message=f"Unable to add owners to pages {', '.join(page_id_ls)}",
            res=res,
        )

    res.response = f"Successfully added owners to pages {', '.join(page_id_ls)}"

    return res



from __future__ import annotations

import datetime as dt
import uuid
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable

import httpx

from ...auth import DomoAuth
from ...base import (
    DomoEntity_w_Lineage,
    DomoEnumMixin,
    entities as dmen,
    exceptions as dmde,
)
from ...routes import publish as publish_routes
from ...utils import chunk_execution as dmce
from ..subentity import lineage as dmdl

__all__ = [
    "DomoPublication_Content_Enum",
    "DomoPublication_Content",
    "DomoPublication_UnexpectedContentType",
    "DomoPublication",
    "DomoSubscription_NoParentAuth",
    "DomoSubscription_NoParent",
    "DomoSubscription",
    "DomoEverywhere",
]


class DomoSubscription_NoParentAuth(dmde.ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            cls_instance=cls_instance,
            entity_id="subscription_id",
            message="must pass parent_auth or parent_auth_retrieval_fn which returns an instance of auth given self",
        )


class DomoSubscription_NoParent(dmde.ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            cls_instance=cls_instance,
            entity_id="subscription_id",
            message="unable to retrieve parent publication",
        )


class DomoPublication_Content_Enum(DomoEnumMixin, Enum):
    from .. import (
        DomoAppStudio as dmas,
        DomoCard as dmac,
        DomoDataset as dmds,
        DomoPage as dmpg,
    )

    CARD = dmac.DomoCard
    DATASET = dmds.DomoDataset
    DATA_APP = dmas.DomoAppStudio
    PAGE = dmpg.DomoPage


@dataclass
class DomoPublication_Content:
    auth: DomoAuth

    content_id: str
    entity_type: str
    entity_id: str
    entity_domain: str
    is_v2: bool
    is_direct_content: bool

    created_dt: dt.datetime
    updated_dt: dt.datetime = None

    subscriber_content_id: str = None
    subscriber_insance: str = None

    entity: Any = field(repr=False, default=None)
    parent: Any = field(repr=False, default=None)

    """the publication content is the content from the publisher instance that is being distributed to subscribers"""

    @classmethod
    def from_dict(cls, obj: dict, auth: DomoAuth, parent: Any = None):
        entity_type = obj.get("content").get("type")
        return cls(
            auth=auth,
            content_id=obj["id"],
            entity_id=obj.get("content").get("domoObjectId"),
            entity_domain=obj.get("content").get("domain"),
            is_v2=obj.get("isV2"),
            created_dt=(
                dt.datetime.fromtimestamp(obj["created"] / 1000)
                if obj["created"]
                else None
            ),
            updated_dt=(
                dt.datetime.fromtimestamp(obj.get("content").get("updated") / 1000)
                if obj.get("content").get("updated")
                else None
            ),
            is_direct_content=obj.get("useDirectContent"),
            parent=parent,
            entity_type=entity_type,
            entity=DomoPublication_Content_Enum[entity_type].value,
        )

    async def get_entity(
        self, debug_api: bool = False, session: httpx.AsyncClient = None
    ):
        """get the entity from the publication content"""
        if not self.entity:
            self.entity = DomoPublication_Content_Enum[self.entity_type].value

        self.entity = await self.entity.get_entity_by_id(
            auth=self.auth,
            entity_id=self.entity_id,
            debug_api=debug_api,
            session=session,
        )

        return self.entity

    def to_api_json(self):
        return {
            "domain": self.entity_domain,
            "domoObjectId": self.entity_id,
            "customerId": self.entity_domain,
            "type": self.entity_type,
        }


class DomoPublication_UnexpectedContentType(dmde.ClassError):
    def __init__(self, cls_instance, content_type):
        super().__init__(
            cls_instance=cls_instance,
            message=f"DomoPublication_Instantiation: Unexpected content type {content_type}",
        )


@dataclass
class DomoPublication(DomoEntity_w_Lineage):
    name: str
    description: str
    is_v2: bool
    created_dt: dt.datetime

    updated_dt: dt.datetime = None

    subscriptions: list[DomoSubscription] = None

    content: list[DomoPublication_Content] = None

    # content_page_id_ls: list[str] = default = None
    # content_dataset_id_ls: list[str] = field(default_factory=list)
    # content_data_app_id_ls: list[str] = field(default_factory=list)

    @property
    def entity_type(self):
        return "PUBLICATION"

    def __post_init__(self):
        self.Lineage = dmdl.DomoLineage_Publication.from_parent(
            parent=self, auth=self.auth
        )

    def _generate_subscriptions(self, subscription_authorizations_ls, auth):
        self.subscriptions = [
            DomoSubscription.from_dict(obj=sub, auth=auth, parent_publication=self)
            for sub in subscription_authorizations_ls
        ]

    def _generate_content(self, children_ls):
        self.content = [
            DomoPublication_Content.from_dict(child, auth=self.auth, parent=self)
            for child in children_ls
        ]

        return self.content

    @classmethod
    def from_dict(cls, obj, auth: DomoAuth):
        domo_pub = cls(
            id=obj["id"],
            name=obj["name"],
            description=obj["description"],
            created_dt=(
                dt.datetime.fromtimestamp(obj["created"] / 1000)
                if obj["created"]
                else None
            ),
            updated_dt=(
                dt.datetime.fromtimestamp(obj.get("content").get("updated") / 1000)
                if obj.get("content").get("updated")
                else None
            ),
            is_v2=obj["isV2"],
            auth=auth,
            raw=obj,
        )

        if (
            obj.get("subscriptionAuthorizations")
            and len(obj.get("subscriptionAuthorizations")) > 0
        ):
            domo_pub._generate_subscriptions(
                subscription_authorizations_ls=obj["subscriptionAuthorizations"],
                auth=auth,
            )

        if obj.get("children") and len(obj.get("children")) > 0:
            domo_pub._generate_content(obj["children"])

        return domo_pub

    @classmethod
    async def get_by_id(
        cls,
        publication_id,
        auth: DomoAuth,
        return_raw: bool = False,
        timeout=10,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        res = await publish_routes.get_publication_by_id(
            auth=auth,
            publication_id=publication_id,
            timeout=timeout,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
            parent_class=cls.__name__,
        )

        if return_raw:
            return res

        return cls.from_dict(obj=res.response, auth=auth)

    @classmethod
    async def get_entity_by_id(cls, entity_id, **kwargs):
        return await cls.get_by_id(publication_id=entity_id, **kwargs)

    def display_url(self):
        return f"https://{self.auth.domo_instance}.domo.com/admin/domo-everywhere/publications/details?id={self.id}"

    async def get_content_details(
        self,
        subscriber_domain: str,  # must include .domo.com
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        if not subscriber_domain.lower().endswith(".domo.com"):
            subscriber_domain = f"{subscriber_domain}.domo.com"

        res = await publish_routes.get_subscriber_content_details(
            auth=self.auth,
            publication_id=self.id,
            subscriber_instance=subscriber_domain,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            debug_api=debug_api,
            session=session,
            parent_class=self.__class__.__name__,
        )

        publication_content = self.content

        for content in publication_content:
            subscriber_obj = next(
                (
                    subscriber_obj
                    for subscriber_obj in res.response
                    if subscriber_obj["publisherObjectId"] == content.entity_id
                    and subscriber_obj["contentType"] == content.entity_type
                ),
                None,
            )
            if subscriber_obj is not None:
                content.subscriber_content_id = subscriber_obj["subscriberObjectId"]
                content.subscriber_insance = subscriber_obj["subscriberDomain"]

        self.content = publication_content

        return res

    # REVIEW THIS PART DUPLICTED FUNCTION

    # async def get_content_details(
    #     self,
    #     subscriber_domain: str = None,
    #     debug_api: bool = False,
    #     session: httpx.AsyncClient = None,
    #     debug_num_stacks_to_drop: int = 2,
    # ):

    #     print(self)
    #     print("2-------")
    #     res = await publish_routes.get_publication_by_id(
    #         auth=self.auth,
    #         publication_id=self.id,
    #         debug_api=debug_api,
    #         session=session,
    #         debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    #         parent_class=self.__class__.__name__,
    #     )

    #     # Extract subscription details
    #     if subscriber_domain:
    #         subscription_auth = next(
    #             (
    #                 sub
    #                 for sub in res.response.get("subscriptionAuthorizations", [])
    #                 if sub.get("subscriberDomain") == subscriber_domain
    #             ),
    #             None,
    #         )
    #         # if subscription_auth:
    #         #     res.response = subscription_auth
    #         # else:
    #         #     res.response = []

    #     return res

    async def get_publication_entity_by_subscriber_entity(
        self,
        subscriber_domain: str,
        subscriber: Any,  # DomoPage, DomoCard, DomoDataset
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
        is_suppress_errors: bool = False,
    ) -> Any:  # DomoCard, DomoDataset, DomoPage
        res = await self.get_content_details(
            subscriber_domain=subscriber_domain,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        obj = next(
            (row for row in res.response if row["subscriberObjectId"] == subscriber.id),
            None,
        )

        if not obj:
            if not is_suppress_errors:
                pass
                # raise DomoError(
                #         =self,
                #         message=f"get_publication_entity_by_subscriber_entity: No matching publication content found for subscriber id {subscriber.id} in publication id {self.id}",
                #     )
            return None

        # Resolve publisher entity. If the provided 'subscriber' is a FederatedDomoDataset
        # we should fetch the publisher-side dataset as the default dataset class (not the federated wrapper).
        publisher_obj_id = obj["publisherObjectId"]

        # Lazy import to avoid circular imports
        if subscriber.__class__.__name__ == "FederatedDomoDataset":
            try:
                from ..DomoDataset.core import DomoDataset as PublisherDataset

                return await PublisherDataset.get_by_id(
                    dataset_id=publisher_obj_id,
                    auth=self.auth,
                    is_use_default_dataset_class=True,
                )
            except Exception:
                # Fall back to calling the subscriber's resolver if publisher fetch fails
                return await subscriber.get_entity_by_id(
                    entity_id=publisher_obj_id, auth=self.auth
                )

        # Default behavior: delegate to subscriber's get_entity_by_id
        return await subscriber.get_entity_by_id(
            entity_id=publisher_obj_id, auth=self.auth
        )

    @classmethod
    async def create_publication(
        cls,
        auth: DomoAuth,
        name: str,
        content_ls: list[DomoPublication_Content],
        subscription_ls: list[DomoSubscription],
        unique_id: str = None,
        description: str = None,
        debug_api: bool = False,
    ):
        if not isinstance(subscription_ls, list):
            subscription_ls = [subscription_ls]

        domain_ls = []
        content_json_ls = []
        for sub in subscription_ls:
            domain_ls.append(sub.subscriber_domain)
        for content_item in content_ls:
            content_json_ls.append(content_item.to_api_json())

        unique_id = unique_id or str(uuid.uuid4())

        body = publish_routes.generate_publish_body(
            url=f"{auth.domo_instance}.domo.com",
            sub_domain_ls=domain_ls,
            content_ls=content_json_ls,
            name=name,
            unique_id=unique_id,
            description=description or "",
            is_new=True,
        )

        res = await publish_routes.create_publish_job(
            auth=auth, body=body, debug_api=debug_api
        )

        return cls.from_dict(obj=res.response, auth=auth)

    async def revoke_subscription_auth(
        self,
        auth: DomoAuth = None,
        subscription_id: str = None,
        subscription: DomoSubscription = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
    ):
        # Note: This function may not have a direct route implementation
        # It would need to be implemented based on the available API endpoints
        auth = auth or self.auth
        subscription_id = subscription_id or subscription.id

        # This is a placeholder implementation - would need actual API route
        raise NotImplementedError("revoke_subscription_auth route not yet implemented")

    async def update_publication(
        self,
        auth: DomoAuth = None,
        content_ls: list[DomoPublication_Content] = None,
        description: str = None,
        name: str = None,
        subscription_ls: list[DomoSubscription] = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
    ):
        auth = auth or self.auth

        # Use the actual available route
        if not isinstance(subscription_ls, list) and subscription_ls:
            subscription_ls = [subscription_ls]

        domain_ls = []
        content_json_ls = []

        if subscription_ls:
            for sub in subscription_ls:
                domain_ls.append(sub.subscriber_domain)

        if content_ls:
            for content_item in content_ls:
                content_json_ls.append(content_item.to_api_json())

        body = publish_routes.generate_publish_body(
            url=f"{auth.domo_instance}.domo.com",
            sub_domain_ls=domain_ls,
            content_ls=content_json_ls,
            name=name or self.name,
            unique_id=self.id,
            description=description or self.description,
            is_new=False,
        )

        res = await publish_routes.update_publish_job(
            auth=auth, publication_id=self.id, body=body
        )

        return res


@dataclass
class DomoSubscription(dmen.DomoEntity):
    """the subscriber represents a location a publication is sent to"""

    id: str
    publication_id: str
    subscriber_domain: str
    publisher_domain: str
    parent_publication: DomoPublication = field(repr=False, default=None)
    created_dt: dt.datetime | None = None

    @classmethod
    def from_dict(cls, obj, auth: DomoAuth, parent_publication: Any = None):
        return cls(
            auth=auth,
            id=obj.get("id") or obj.get("subscriptionId"),
            publication_id=obj["publicationId"],
            subscriber_domain=obj.get("domain") or obj.get("subscriberDomain"),
            publisher_domain=obj.get("publisherDomain"),
            created_dt=(
                (dt.datetime.fromtimestamp(obj.get("created") / 1000))
                if obj.get("created")
                else None
            ),
            raw=obj,
            parent_publication=parent_publication,
        )

    @property
    def entity_type(self):
        return "SUBSCRIPTION"

    def display_url(self):
        return f"https://{self.auth.domo_instance}.domo.com/admin/domo-everywhere/subscriptions"

    @classmethod
    async def get_by_id(
        cls,
        auth: DomoAuth,
        subscription_id: str,
        return_raw: bool = False,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
    ):
        res = await publish_routes.get_subscription_by_id(
            auth=auth,
            subscription_id=subscription_id,
            debug_api=debug_api,
            session=session,
            parent_class=cls.__name__,
        )

        if return_raw:
            return res

        return cls.from_dict(obj=res.response, auth=auth)

    @classmethod
    async def get_entity_by_id(cls, entity_id: str, auth: DomoAuth, **kwargs):
        return await cls.get_by_id(dataset_id=entity_id, auth=auth, **kwargs)

    async def get_parent_publication(
        self,
        parent_auth: DomoAuth = None,
        parent_auth_retrieval_fn: Callable = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
    ):
        if not parent_auth and parent_auth_retrieval_fn:
            parent_auth = parent_auth_retrieval_fn(self)

        if not parent_auth:
            raise DomoSubscription_NoParentAuth(self)

        self.parent_publication = await DomoPublication.get_by_id(
            publication_id=self.publication_id,
            auth=parent_auth,
            debug_api=debug_api,
            session=session,
        )

        return self.parent_publication

    async def get_content_details(
        self,
        parent_auth: DomoAuth = None,
        parent_auth_retrieval_fn: Callable = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        if not self.parent_publication:
            await self.get_parent_publication(
                parent_auth=parent_auth,
                parent_auth_retrieval_fn=parent_auth_retrieval_fn,
                debug_api=debug_api,
                session=session,
            )

        if not self.parent_publication:
            raise DomoSubscription_NoParent(self)

        publication_content = self.parent_publication.content

        res = await publish_routes.get_subscriber_content_details(
            auth=self.parent_publication.auth,
            publication_id=self.publication_id,
            subscriber_instance=self.subscriber_domain,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            debug_api=debug_api,
            session=session,
            parent_class=self.__class__.__name__,
        )

        for content in publication_content:
            subscriber_obj = next(
                (
                    subscriber_obj
                    for subscriber_obj in res.response
                    if subscriber_obj["publisherObjectId"] == content.entity_id
                    and subscriber_obj["contentType"] == content.entity_type
                ),
                None,
            )
            if subscriber_obj is not None:
                content.subscriber_content_id = subscriber_obj["subscriberObjectId"]
                content.subscriber_insance = subscriber_obj["subscriberDomain"]

        self.content = publication_content

        return self.content


@dataclass
class DomoEverywhere:
    auth: DomoAuth = field(repr=False)

    publications: list[DomoPublication] = field(default=None)

    subscriptions: list[DomoSubscription] = field(default=None)

    invitations: list[dict] = field(default=None)

    async def get_publications(
        self,
        search_term: str = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        res = await publish_routes.search_publications(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            search_term=search_term,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        if return_raw:
            return res

        self.publications = await dmce.gather_with_concurrency(
            n=60,
            *[
                DomoPublication.get_by_id(publication_id=obj.get("id"), auth=self.auth)
                for obj in res.response
            ],
        )
        return self.publications

    async def search_publications(
        self,
        search_term: str = None,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        res = await self.get_publications(
            search_term=search_term,
            session=session,
            debug_api=debug_api,
            return_raw=return_raw,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        return res

    async def get_subscriptions(
        self,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_api: bool = False,
    ):
        """get instances subscription summaries"""

        self.subscriptions = []

        res = await publish_routes.get_subscription_summaries(
            auth=self.auth, session=session, debug_api=debug_api
        )

        if return_raw:
            return res

        for sub in res.response:
            domo_sub = DomoSubscription.from_dict(sub, auth=self.auth)

            if sub in self.subscriptions:
                continue
            self.subscriptions.append(domo_sub)

        return self.subscriptions

    async def get_subscription_invitations(
        self, debug_api: bool = False, session: httpx.AsyncClient = None
    ):
        res = await publish_routes.get_subscription_invitations(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            parent_class=self.__class__.__name__,
        )

        self.invitations = res.response

        return res

    async def accept_invite_by_id(
        self,
        subscription_id: str,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
    ):
        res = await publish_routes.accept_invite_by_id(
            auth=self.auth,
            subscription_id=subscription_id,
            debug_api=debug_api,
            session=session,
        )

        if res.status == 200:
            return res.response
        else:
            return None
